#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import argparse
import os
from typing import Dict, List

import pandas as pd
import torch
from pydantic import Field
from torch.autograd import Variable
from torch.utils.data import DataLoader

from ...algorithms.core import (
    ConfigurablePropertyAlgorithmConfiguration,
    Predictor,
    PredictorAlgorithm,
)
from ...frameworks.cgcnn.data import AtomCustomJSONInitializer, CIFData, collate_pool
from ...frameworks.cgcnn.model import CrystalGraphConvNet, Normalizer
from ...frameworks.crystals_rfc.feature_engine import Features
from ...frameworks.crystals_rfc.rf_classifier import RFC
from ..core import DomainSubmodule, S3Parameters


class S3ParametersCrystals(S3Parameters):
    domain: DomainSubmodule = DomainSubmodule("crystals")


class CGCNNParameters(S3ParametersCrystals):
    algorithm_name: str = "cgcnn"
    batch_size: int = Field(description="Prediction batch size", default=256)
    workers: int = Field(description="Number of data loading workers", default=0)


class MetalNonMetalClassifierParameters(S3ParametersCrystals):
    algorithm_name: str = "RFC"
    algorithm_application: str = "MetalNonMetalClassifier"


class FormationEnergyParameters(CGCNNParameters):
    algorithm_application: str = "FormationEnergy"


class AbsoluteEnergyParameters(CGCNNParameters):
    algorithm_application: str = "AbsoluteEnergy"


class BandGapParameters(CGCNNParameters):
    algorithm_application: str = "BandGap"


class FermiEnergyParameters(CGCNNParameters):
    algorithm_application: str = "FermiEnergy"


class BulkModuliParameters(CGCNNParameters):
    algorithm_application: str = "BulkModuli"


class ShearModuliParameters(CGCNNParameters):
    algorithm_application: str = "ShearModuli"


class PoissonRatioParameters(CGCNNParameters):
    algorithm_application: str = "PoissonRatio"


class MetalSemiconductorClassifierParameters(CGCNNParameters):
    algorithm_application: str = "MetalSemiconductorClassifier"


class _CGCNN(PredictorAlgorithm):
    """Base class for all cgcnn-based predictive algorithms."""

    def __init__(self, parameters: CGCNNParameters):

        # Set up the configuration from the parameters
        configuration = ConfigurablePropertyAlgorithmConfiguration(
            algorithm_type=parameters.algorithm_type,
            domain=parameters.domain,
            algorithm_name=parameters.algorithm_name,
            algorithm_application=parameters.algorithm_application,
            algorithm_version=parameters.algorithm_version,
        )

        self.batch_size = parameters.batch_size
        self.workers = parameters.workers

        # The parent constructor calls `self.get_model`.
        super().__init__(configuration=configuration)

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """

        existing_models = os.listdir(resources_path)
        existing_models = [
            file for file in existing_models if file.endswith(".pth.tar")
        ]

        if len(existing_models) > 1:
            raise ValueError(
                "Only one model should be located in the specified model path."
            )
        elif len(existing_models) == 0:
            raise ValueError("Model does not exist in the specified model path.")

        model_path = os.path.join(resources_path, existing_models[0])

        checkpoint = torch.load(model_path, map_location=lambda storage, loc: storage)

        model_args = argparse.Namespace(**checkpoint["args"])

        normalizer = Normalizer(torch.zeros(3))
        normalizer.load_state_dict(checkpoint["normalizer"])

        atom_initialization = AtomCustomJSONInitializer(
            os.path.join(resources_path, "atom_init.json")
        )

        # Wrapper to get the predictions
        def informative_model(cif_path: str) -> Dict[str, List[float]]:

            dataset = CIFData(cif_path, atom_initialization=atom_initialization)
            test_loader = DataLoader(
                dataset,
                batch_size=self.batch_size,
                num_workers=self.workers,
                collate_fn=collate_pool,
            )

            # build model
            structures, _, _ = dataset[0]
            orig_atom_fea_len = structures[0].shape[-1]
            nbr_fea_len = structures[1].shape[-1]  # type: ignore

            model = CrystalGraphConvNet(
                orig_atom_fea_len,
                nbr_fea_len,
                atom_fea_len=model_args.atom_fea_len,
                n_conv=model_args.n_conv,
                h_fea_len=model_args.h_fea_len,
                n_h=model_args.n_h,
                classification=True if model_args.task == "classification" else False,
            )

            model.load_state_dict(checkpoint["state_dict"])
            model.eval()

            test_preds = []
            test_cif_ids = []

            for input, _, batch_cif_ids in test_loader:
                with torch.no_grad():
                    input_var = (
                        Variable(input[0]),
                        Variable(input[1]),
                        input[2],
                        input[3],
                    )

                # compute output
                output = model(*input_var)

                # record loss
                if model_args.task == "classification":
                    test_pred = torch.exp(output.data.cpu())
                    test_preds += test_pred[:, 1].tolist()
                else:
                    test_pred = normalizer.denorm(output.data.cpu())
                    test_preds += test_pred.view(-1).tolist()
                test_cif_ids += batch_cif_ids

            return {"cif_ids": test_cif_ids, "predictions": test_preds}  # type: ignore

        return informative_model


class MetalNonMetalClassifier(PredictorAlgorithm):
    """Metal/non-metal classifier class."""

    def __init__(self, parameters: MetalNonMetalClassifierParameters):

        # Set up the configuration from the parameters
        configuration = ConfigurablePropertyAlgorithmConfiguration(
            algorithm_type=parameters.algorithm_type,
            domain=parameters.domain,
            algorithm_name=parameters.algorithm_name,
            algorithm_application=parameters.algorithm_application,
            algorithm_version=parameters.algorithm_version,
        )

        # The parent constructor calls `self.get_model`.
        super().__init__(configuration=configuration)

    def get_model(self, resources_path: str) -> Predictor:
        """Instantiate the actual model.

        Args:
            resources_path: local path to model files.

        Returns:
            Predictor: the model.
        """

        rfc = RFC()

        rfc.load_model(resources_path)

        # Wrapper to get the predictions
        def informative_model(formula_file: str) -> Dict[str, List[str]]:
            # getting the feature
            feature_eng = Features(formula_file=formula_file)
            features = feature_eng.get_features()

            # getting the targets and symmetries
            df = pd.DataFrame(features)
            pred_x = df.iloc[:, 1:].values

            # getting the chemical formulas
            df_mat = pd.read_csv(formula_file, header=None)
            formulas = df_mat.iloc[:, 0].to_list()

            predictions = rfc.predict(pred_x=pred_x)

            return {"formulas": formulas, "predictions": predictions}

        return informative_model


class FormationEnergy(_CGCNN):
    @classmethod
    def get_description(cls) -> str:
        text = """
        This model predicts the formation energy per atom using the CGCNN framework.
        For more details see: https://doi.org/10.1103/PhysRevLett.120.145301.
        """
        return text


class AbsoluteEnergy(_CGCNN):
    @classmethod
    def get_description(cls) -> str:
        text = """
        This model predicts the absolute energy of crystals using the CGCNN framework.
        For more details see: https://doi.org/10.1103/PhysRevLett.120.145301.
        """
        return text


class BandGap(_CGCNN):
    @classmethod
    def get_description(cls) -> str:
        text = """
        This model predicts the band gap of crystals using the CGCNN framework.
        For more details see: https://doi.org/10.1103/PhysRevLett.120.145301.
        """
        return text


class FermiEnergy(_CGCNN):
    @classmethod
    def get_description(cls) -> str:
        text = """
        This model predicts the Fermi energy of crystals using the CGCNN framework.
        For more details see: https://doi.org/10.1103/PhysRevLett.120.145301.
        """
        return text


class BulkModuli(_CGCNN):
    @classmethod
    def get_description(cls) -> str:
        text = """
        This model predicts the bulk moduli of crystals using the CGCNN framework.
        For more details see: https://doi.org/10.1103/PhysRevLett.120.145301.
        """
        return text


class ShearModuli(_CGCNN):
    @classmethod
    def get_description(cls) -> str:
        text = """
         This model predicts the shear moduli of crystals using the CGCNN framework.
        For more details see: https://doi.org/10.1103/PhysRevLett.120.145301.
         """
        return text


class PoissonRatio(_CGCNN):
    @classmethod
    def get_description(cls) -> str:
        text = """
        This model predicts the Poisson ratio of crystals using the CGCNN framework.
        For more details see: https://doi.org/10.1103/PhysRevLett.120.145301.
         """
        return text


class MetalSemiconductorClassifier(_CGCNN):
    @classmethod
    def get_description(cls) -> str:
        text = """
         This model predicts whether a given crystal is metal or semiconductor using the CGCNN framework.
        For more details see: https://doi.org/10.1103/PhysRevLett.120.145301.
         """
        return text



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
from copy import deepcopy

import torch
from torch_geometric.data import Batch, Data
from torch_geometric.nn import radius, radius_graph
from torch_geometric.utils import dense_to_sparse, to_dense_adj
from torch_scatter import scatter_add
from torch_sparse import coalesce


def assemble_atom_pair_feature(node_attr, edge_index, edge_attr):
    h_row, h_col = node_attr[edge_index[0]], node_attr[edge_index[1]]
    h_pair = torch.cat([h_row * h_col, edge_attr], dim=-1)  # (E, 2H)
    return h_pair


def _extend_graph_order(num_nodes, edge_index, edge_type, order=3):
    """
    Args:
        num_nodes:  Number of atoms.
        edge_index: Bond indices of the original graph.
        edge_type:  Bond types of the original graph.
        order:  Extension order.
    Returns:
        new_edge_index: Extended edge indices. new_edge_type: Extended edge types.
    """

    def binarize(x):
        return torch.where(x > 0, torch.ones_like(x), torch.zeros_like(x))

    def get_higher_order_adj_matrix(adj, order):
        """
        Args:
            adj:        (N, N)
            type_mat:   (N, N)
        Returns:
            Following attributes will be updated:
              - edge_index
              - edge_type
            Following attributes will be added to the data object:
              - bond_edge_index: Original edge_index.
        """
        adj_mats = [
            torch.eye(adj.size(0), dtype=torch.long, device=adj.device),
            binarize(adj + torch.eye(adj.size(0), dtype=torch.long, device=adj.device)),
        ]

        for i in range(2, order + 1):
            adj_mats.append(binarize(adj_mats[i - 1] @ adj_mats[1]))
        order_mat = torch.zeros_like(adj)

        for i in range(1, order + 1):
            order_mat += (adj_mats[i] - adj_mats[i - 1]) * i

        return order_mat

    num_types = 22
    # given from len(BOND_TYPES), where BOND_TYPES = {t: i for i, t in enumerate(BT.names.values())}
    # from rdkit.Chem.rdchem import BondType as BT
    N = num_nodes
    adj = to_dense_adj(edge_index).squeeze(0)
    adj_order = get_higher_order_adj_matrix(adj, order)  # (N, N)

    type_mat = to_dense_adj(edge_index, edge_attr=edge_type).squeeze(0)  # (N, N)
    type_highorder = torch.where(
        adj_order > 1, num_types + adj_order - 1, torch.zeros_like(adj_order)
    )
    assert (type_mat * type_highorder == 0).all()
    type_new = type_mat + type_highorder

    new_edge_index, new_edge_type = dense_to_sparse(type_new)
    _, edge_order = dense_to_sparse(adj_order)

    # data.bond_edge_index = data.edge_index  # Save original edges
    new_edge_index, new_edge_type = coalesce(
        new_edge_index, new_edge_type.long(), N, N
    )  # modify data

    return new_edge_index, new_edge_type


def _extend_to_radius_graph(
    pos,
    edge_index,
    edge_type,
    cutoff,
    batch,
    unspecified_type_number=0,
    is_sidechain=None,
):
    assert edge_type.dim() == 1
    N = pos.size(0)

    bgraph_adj = torch.sparse.LongTensor(edge_index, edge_type, torch.Size([N, N]))  # type: ignore

    if is_sidechain is None:
        rgraph_edge_index = radius_graph(pos, r=cutoff, batch=batch)  # (2, E_r)
    else:
        # fetch sidechain and its batch index
        is_sidechain = is_sidechain.bool()
        dummy_index = torch.arange(pos.size(0), device=pos.device)
        sidechain_pos = pos[is_sidechain]
        sidechain_index = dummy_index[is_sidechain]
        sidechain_batch = batch[is_sidechain]

        assign_index = radius(
            x=pos, y=sidechain_pos, r=cutoff, batch_x=batch, batch_y=sidechain_batch
        )
        r_edge_index_x = assign_index[1]
        r_edge_index_y = assign_index[0]
        r_edge_index_y = sidechain_index[r_edge_index_y]

        rgraph_edge_index1 = torch.stack((r_edge_index_x, r_edge_index_y))  # (2, E)
        rgraph_edge_index2 = torch.stack((r_edge_index_y, r_edge_index_x))  # (2, E)
        rgraph_edge_index = torch.cat(
            (rgraph_edge_index1, rgraph_edge_index2), dim=-1
        )  # (2, 2E)
        # delete self loop
        rgraph_edge_index = rgraph_edge_index[
            :, (rgraph_edge_index[0] != rgraph_edge_index[1])
        ]

    rgraph_adj = torch.sparse.LongTensor(  # type: ignore
        rgraph_edge_index,
        torch.ones(rgraph_edge_index.size(1)).long().to(pos.device)
        * unspecified_type_number,
        torch.Size([N, N]),
    )

    composed_adj = (bgraph_adj + rgraph_adj).coalesce()  # Sparse (N, N, T)

    new_edge_index = composed_adj.indices()
    new_edge_type = composed_adj.values().long()

    return new_edge_index, new_edge_type


def extend_graph_order_radius(
    num_nodes,
    pos,
    edge_index,
    edge_type,
    batch,
    order=3,
    cutoff=10.0,
    extend_order=True,
    extend_radius=True,
    is_sidechain=None,
):
    if extend_order:
        edge_index, edge_type = _extend_graph_order(
            num_nodes=num_nodes, edge_index=edge_index, edge_type=edge_type, order=order
        )

    if extend_radius:
        edge_index, edge_type = _extend_to_radius_graph(
            pos=pos,
            edge_index=edge_index,
            edge_type=edge_type,
            cutoff=cutoff,
            batch=batch,
            is_sidechain=is_sidechain,
        )

    return edge_index, edge_type


def get_distance(pos, edge_index):
    return (pos[edge_index[0]] - pos[edge_index[1]]).norm(dim=-1)


def graph_field_network(score_d, pos, edge_index, edge_length):
    """
    Transformation to make the epsilon predicted from the diffusion model roto-translational equivariant. See equations
    5-7 of the GeoDiff Paper https://arxiv.org/pdf/2203.02923.pdf
    """
    N = pos.size(0)
    dd_dr = (1.0 / edge_length) * (pos[edge_index[0]] - pos[edge_index[1]])  # (E, 3)
    score_pos = scatter_add(
        dd_dr * score_d, edge_index[0], dim=0, dim_size=N
    ) + scatter_add(
        -dd_dr * score_d, edge_index[1], dim=0, dim_size=N
    )  # (N, 3)
    return score_pos


def clip_norm(vec, limit, p=2):
    norm = torch.norm(vec, dim=-1, p=2, keepdim=True)
    denom = torch.where(norm > limit, limit / norm, torch.ones_like(norm))
    return vec * denom


def is_local_edge(edge_type):
    return edge_type > 0


def repeat_data(data: Data, num_repeat: int) -> Batch:
    """
    Args:
        data:  An `torch_geometric.data.Data` object.

    Returns:
        batch: A copy of `data` repetead `num_repeat` times.
    """
    datas = [copy.deepcopy(data) for i in range(num_repeat)]
    return Batch.from_data_list(datas)


def repeat_batch(batch: Batch, num_repeat: int) -> Batch:
    """
    Args:
        batch:  An `torch_geometric.data.Batch` object.

    Returns:
        batch: A copy of `batch` repetead `num_repeat` times.
    """
    datas = batch.to_data_list()
    new_data = []
    for i in range(num_repeat):
        new_data += copy.deepcopy(datas)
    return Batch.from_data_list(new_data)


def set_rdmol_positions(rdkit_mol, pos):
    """
    Args:
        rdkit_mol:  An `rdkit.Chem.rdchem.Mol` object.
        pos: (N_atoms, 3)

    Returns:
        mol: A copy of `rdkit_mol` with the positions set to `pos`.
    """
    mol = deepcopy(rdkit_mol)
    set_rdmol_positions_(mol, pos)
    return mol


def set_rdmol_positions_(mol, pos):
    """
    Args:
        rdkit_mol:  An `rdkit.Chem.rdchem.Mol` object.
        pos: (N_atoms, 3)

    Returns:
        mol: A copy of `rdkit_mol` with the positions set to `pos`.
    """
    for i in range(pos.shape[0]):
        mol.GetConformer(0).SetAtomPosition(i, pos[i].tolist())
    return mol



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
from dataclasses import dataclass
from typing import Callable, List, Union

import numpy as np
import torch
import torch.nn.functional as F
from diffusers.utils import BaseOutput
from torch import Tensor, nn
from torch.nn import Embedding, Linear, Module, ModuleList, Sequential
from torch_geometric.nn import MessagePassing
from torch_geometric.typing import Adj, OptPairTensor, OptTensor, Size
from torch_sparse import SparseTensor

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@dataclass
class MoleculeGNNOutput(BaseOutput):
    """Hidden states output. Output of last layer of model."""

    sample: torch.FloatTensor


class MultiLayerPerceptron(nn.Module):
    """Multi-layer Perceptron. Note there is no activation or dropout in the last layer."""

    def __init__(
        self,
        input_dim: int,
        hidden_dims: List[int],
        activation: str = "relu",
        dropout: float = 0,
    ) -> None:
        """Initialize multi-layer perceptron.

        Args:
            input_dim: input dimension
            hidden_dim: hidden dimensions
            activation: activation function
            dropout: dropout rate
        """
        super(MultiLayerPerceptron, self).__init__()

        self.dims = [input_dim] + hidden_dims
        if isinstance(activation, str):
            self.activation = getattr(F, activation)
        else:
            logger.info(
                f"Warning, activation passed {activation} is not string and ignored"
            )
            self.activation = None
        if dropout > 0:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = None  # type: ignore

        self.layers = nn.ModuleList()
        for i in range(len(self.dims) - 1):
            self.layers.append(nn.Linear(self.dims[i], self.dims[i + 1]))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            x: input tensor of shape (batch_size, input_dim)

        Returns:
            output mlp.
        """
        for i, layer in enumerate(self.layers):
            x = layer(x)
            if i < len(self.layers) - 1:
                if self.activation:
                    x = self.activation(x)
                if self.dropout:
                    x = self.dropout(x)
        return x


class ShiftedSoftplus(torch.nn.Module):
    """Shifted softplus activation function."""

    def __init__(self) -> None:
        super(ShiftedSoftplus, self).__init__()
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return F.softplus(x) - self.shift


class CFConv(MessagePassing):
    """CFConv layer."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        num_filters: int,
        mlp: Callable,
        cutoff: float,
        smooth: bool,
    ) -> None:
        """Construct a CFConv layer.

        Args:
            in_channels: size of each input.
            out_channels: size of each output.
            num_filters: number of filters.
            mlp: mlp hidden dimensions.
            cutoff: cutoff distance.
            smooth: whether to use smooth cutoff.
        """
        super(CFConv, self).__init__(aggr="add")
        self.lin1 = Linear(in_channels, num_filters, bias=False)
        self.lin2 = Linear(num_filters, out_channels)
        self.nn = mlp
        self.cutoff = cutoff
        self.smooth = smooth

        self.reset_parameters()

    def reset_parameters(self) -> None:
        """Initialize parameters."""
        torch.nn.init.xavier_uniform_(self.lin1.weight)
        torch.nn.init.xavier_uniform_(self.lin2.weight)
        self.lin2.bias.data.fill_(0)

    def forward(
        self, x: torch.Tensor, edge_index, edge_length, edge_attr
    ) -> torch.Tensor:
        """Forward pass.

        Args:
            x: input tensor.
            edge_index: edge indices.
            edge_length: edge lengths.
            edge_attr: edge attributes.

        Returns:
            output tensor.
        """
        if self.smooth:
            C = 0.5 * (torch.cos(edge_length * np.pi / self.cutoff) + 1.0)
            C = (
                C * (edge_length <= self.cutoff) * (edge_length >= 0.0)
            )  # Modification: cutoff
        else:
            C = (edge_length <= self.cutoff).float()
        W = self.nn(edge_attr) * C.view(-1, 1)

        x = self.lin1(x)
        x = self.propagate(edge_index, x=x, W=W)
        x = self.lin2(x)
        return x

    def message(self, x_j: torch.Tensor, W) -> torch.Tensor:
        return x_j * W


class InteractionBlock(torch.nn.Module):
    """Interaction block."""

    def __init__(
        self,
        hidden_channels: int,
        num_gaussians: int,
        num_filters: int,
        cutoff: float,
        smooth: bool,
    ) -> None:
        """Construct an interaction block.

        Args:
            hidden_channels: number of hidden channels.
            num_gaussians: number of gaussians.
            num_filters: number of filters.
            cutoff: cutoff distance.
            smooth: whether to use smooth cutoff.
        """
        super(InteractionBlock, self).__init__()
        mlp = Sequential(
            Linear(num_gaussians, num_filters),
            ShiftedSoftplus(),
            Linear(num_filters, num_filters),
        )
        self.conv = CFConv(
            hidden_channels, hidden_channels, num_filters, mlp, cutoff, smooth
        )
        self.act = ShiftedSoftplus()
        self.lin = Linear(hidden_channels, hidden_channels)

    def forward(
        self, x: torch.Tensor, edge_index, edge_length, edge_attr
    ) -> torch.Tensor:
        """Forward pass.

        Args:
            x: input tensor.
            edge_index: edge indices.
            edge_length: edge lengths.
            edge_attr: edge attributes.

        Returns:
            output tensor.
        """
        x = self.conv(x, edge_index, edge_length, edge_attr)
        x = self.act(x)
        x = self.lin(x)
        return x


class SchNetEncoder(Module):
    """SchNet encoder."""

    def __init__(
        self,
        hidden_channels: int = 128,
        num_filters: int = 128,
        num_interactions: int = 6,
        edge_channels: int = 100,
        cutoff: float = 10.0,
        smooth: bool = False,
    ) -> None:
        """Construct a SchNet encoder.

        Args:
            hidden_channels: number of hidden channels.
            num_filters: number of filters.
            num_interactions: number of interactions.
            edge_channels: number of edge channels.
            cutoff: cutoff distance.
            smooth: whether to use smooth cutoff.
        """
        super().__init__()

        self.hidden_channels = hidden_channels
        self.num_filters = num_filters
        self.num_interactions = num_interactions
        self.cutoff = cutoff

        self.embedding = Embedding(100, hidden_channels, max_norm=10.0)

        self.interactions = ModuleList()
        for _ in range(num_interactions):
            block = InteractionBlock(
                hidden_channels, edge_channels, num_filters, cutoff, smooth
            )
            self.interactions.append(block)

    def forward(
        self,
        z: torch.Tensor,
        edge_index: torch.Tensor,
        edge_length: torch.Tensor,
        edge_attr: torch.Tensor,
        embed_node: bool = True,
    ) -> torch.Tensor:
        """Forward pass.

        Args:
            z: input tensor.
            edge_index: edge indices.
            edge_length: edge lengths.
            edge_attr: edge attributes.
            embed_node: whether to embed node.

        Returns:
            output tensor.
        """
        if embed_node:
            assert z.dim() == 1 and z.dtype == torch.long
            h = self.embedding(z)
        else:
            h = z
        for interaction in self.interactions:
            h = h + interaction(h, edge_index, edge_length, edge_attr)

        return h


class GINEConv(MessagePassing):
    """GINECONV layer.
    Custom class of the graph isomorphism operator from the "How Powerful are Graph Neural Networks?
    https://arxiv.org/abs/1810.00826 paper. Note that this implementation has the added option of a custom activation.
    """

    def __init__(
        self,
        mlp: Callable,
        eps: float = 0.0,
        train_eps: bool = False,
        activation: str = "softplus",
        **kwargs,
    ) -> None:
        """Construct a GINEConv layer.

        Args:
            mlp: MLP.
            eps: epsilon.
            train_eps: whether to train epsilon.
            activation: activation function.
        """
        super(GINEConv, self).__init__(aggr="add", **kwargs)
        self.nn = mlp
        self.initial_eps = eps

        if isinstance(activation, str):
            self.activation = getattr(F, activation)
        else:
            self.activation = None

        if train_eps:
            self.eps = torch.nn.Parameter(torch.Tensor([eps]))
        else:
            self.register_buffer("eps", torch.Tensor([eps]))

    def forward(
        self,
        x: Union[Tensor, OptPairTensor],
        edge_index: Adj,
        edge_attr: OptTensor = None,
        size: Size = None,
    ) -> torch.Tensor:
        """Forward pass.

        Args:
            x: input tensor.
            edge_index: edge indices.
            edge_attr: edge attributes.
            size: size.

        Returns:
            output tensor.
        """

        if isinstance(x, torch.Tensor):
            x = (x, x)

        # Node and edge feature dimensionalites need to match.
        if isinstance(edge_index, torch.Tensor):
            assert edge_attr is not None
            assert x[0].size(-1) == edge_attr.size(-1)
        elif isinstance(edge_index, SparseTensor):
            assert x[0].size(-1) == edge_index.size(-1)

        # propagate_type: (x: OptPairTensor, edge_attr: OptTensor)
        out = self.propagate(edge_index, x=x, edge_attr=edge_attr, size=size)

        x_r = x[1]
        if x_r is not None:
            out += (1 + self.eps) * x_r

        return self.nn(out)

    def message(self, x_j: torch.Tensor, edge_attr: torch.Tensor) -> torch.Tensor:
        """Message function.

        Args:
            x_j: input tensor.
            edge_attr: edge attributes.

        Returns:
            message passing aggregation.
        """
        if self.activation:
            return self.activation(x_j + edge_attr)
        else:
            return x_j + edge_attr

    def __repr__(self):
        return "{}(nn={})".format(self.__class__.__name__, self.nn)


class GINEncoder(torch.nn.Module):
    """GIN encoder."""

    def __init__(
        self,
        hidden_dim: int,
        num_convs: int = 3,
        activation: str = "relu",
        short_cut: bool = True,
        concat_hidden: bool = False,
    ) -> None:
        """Construct a GIN encoder.

        Args:
            hidden_dim: number of hidden channels.
            num_convs: number of convolutions.
            activation: activation function.
            short_cut: whether to use short cut.
            concat_hidden: whether to concatenate hidden.
        """
        super().__init__()

        self.hidden_dim = hidden_dim
        self.num_convs = num_convs
        self.short_cut = short_cut
        self.concat_hidden = concat_hidden
        self.node_emb = nn.Embedding(100, hidden_dim)

        if isinstance(activation, str):
            self.activation = getattr(F, activation)
        else:
            self.activation = None

        self.convs = nn.ModuleList()
        for i in range(self.num_convs):
            self.convs.append(
                GINEConv(
                    MultiLayerPerceptron(
                        hidden_dim, [hidden_dim, hidden_dim], activation=activation
                    ),
                    activation=activation,
                )
            )

    def forward(
        self, z: torch.Tensor, edge_index: torch.Tensor, edge_attr: torch.Tensor
    ) -> torch.Tensor:
        """Forward pass.

        args:
            z: input tensor.
            edge_index: edge indices.
            edge_attr: edge attributes.

        returns:
            graph with node feature.
        """

        node_attr = self.node_emb(z)  # (num_node, hidden)

        hiddens = []
        conv_input = node_attr  # (num_node, hidden)

        for conv_idx, conv in enumerate(self.convs):
            hidden = conv(conv_input, edge_index, edge_attr)
            if conv_idx < len(self.convs) - 1 and self.activation is not None:
                hidden = self.activation(hidden)
            assert hidden.shape == conv_input.shape
            if self.short_cut and hidden.shape == conv_input.shape:
                hidden += conv_input

            hiddens.append(hidden)
            conv_input = hidden

        if self.concat_hidden:
            node_feature = torch.cat(hiddens, dim=-1)
        else:
            node_feature = hiddens[-1]

        return node_feature


class MLPEdgeEncoder(Module):
    """MLP edge encoder."""

    def __init__(self, hidden_dim: int = 100, activation: str = "relu") -> None:
        """Construct a MLP edge encoder.

        Args:
            hidden_dim: number of hidden channels.
            activation: activation function.
        """
        super().__init__()
        self.hidden_dim = hidden_dim
        self.bond_emb = Embedding(100, embedding_dim=self.hidden_dim)
        self.mlp = MultiLayerPerceptron(
            1, [self.hidden_dim, self.hidden_dim], activation=activation
        )

    @property
    def out_channels(self):
        return self.hidden_dim

    def forward(
        self, edge_length: torch.Tensor, edge_type: torch.Tensor
    ) -> torch.Tensor:
        """
        Args:
            edge_length: The length of edges.
            edge_type: The type of edges.
        Returns:
            the output representation of edges.
        """
        d_emb = self.mlp(edge_length)  # (num_edge, hidden_dim)
        edge_attr = self.bond_emb(edge_type)  # (num_edge, hidden_dim)
        return d_emb * edge_attr  # (num_edge, hidden)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Train module implementation."""

import logging
from argparse import Namespace
from typing import Any, Dict

import sentencepiece as _sentencepiece
import torch as _torch
import tensorflow as _tensorflow
import pytorch_lightning as pl
from pytorch_lightning.callbacks import ModelCheckpoint
from pytorch_lightning.loggers import TensorBoardLogger

from ..arg_parser.parser import parse_arguments_from_config
from ..dataloader.data_module import GranularDataModule
from ..dataloader.dataset import build_dataset_and_architecture
from ..ml.models import AUTOENCODER_ARCHITECTURES
from ..ml.module import GranularModule

# imports that have to be loaded before lightning to avoid segfaults
_sentencepiece
_tensorflow
_torch

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def train_granular(configuration: Dict[str, Any]) -> None:
    """Train a granular given a configuration.
    Args:
        configuration: a configuration dictionary.
    """
    arguments = Namespace(**configuration)
    datasets = []
    architecture_autoencoders = []
    architecture_latent_models = []
    for model in arguments.model_list:
        logger.info(f"dataset preparation for model={model}")
        hparams = configuration[model]
        model_type = hparams["type"].lower()
        dataset, architecture = build_dataset_and_architecture(
            hparams["name"],
            hparams["data_path"],
            hparams["data_file"],
            hparams["dataset_type"],
            hparams["type"],
            hparams,
        )
        datasets.append(dataset)
        if model_type in AUTOENCODER_ARCHITECTURES:
            architecture_autoencoders.append(architecture)
        else:
            architecture_latent_models.append(architecture)
    dm = GranularDataModule(
        datasets,
        batch_size=getattr(arguments, "batch_size", 64),
        validation_split=getattr(arguments, "validation_split", None),
        validation_indices_file=getattr(arguments, "validation_indices_file", None),
        stratified_batch_file=getattr(arguments, "stratified_batch_file", None),
        stratified_value_name=getattr(arguments, "stratified_value_name", None),
        num_workers=getattr(arguments, "num_workers", 1),
    )
    dm.prepare_data()
    module = GranularModule(
        architecture_autoencoders=architecture_autoencoders,
        architecture_latent_models=architecture_latent_models,
        lr=getattr(arguments, "lr", 0.0001),
        test_output_path=getattr(arguments, "test_output_path", "./test"),
    )
    tensorboard_logger = TensorBoardLogger(
        "logs", name=getattr(arguments, "basename", "default")
    )
    checkpoint_callback = ModelCheckpoint(
        every_n_epochs=getattr(arguments, "checkpoint_every_n_val_epochs", 5),
        save_top_k=-1,
    )
    trainer = pl.Trainer.from_argparse_args(
        arguments,
        profiler="simple",
        logger=tensorboard_logger,
        auto_lr_find=True,
        log_every_n_steps=getattr(arguments, "trainer_log_every_n_steps", 50),
        callbacks=[checkpoint_callback],
        max_epochs=getattr(arguments, "epoch", 1),
    )
    trainer.fit(module, dm)


def train_granular_main() -> None:
    """Train a granular module parsing arguments from config and standard input."""
    train_granular(configuration=vars(parse_arguments_from_config()))



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import Any, Tuple, Union, Optional

import torch
from diffusers.configuration_utils import ConfigMixin
from diffusers.modeling_utils import ModelMixin
from torch import nn
from torch_geometric.data import Data

from .layers import (
    GINEncoder,
    MLPEdgeEncoder,
    MoleculeGNNOutput,
    MultiLayerPerceptron,
    SchNetEncoder,
)
from .utils import (
    assemble_atom_pair_feature,
    clip_norm,
    extend_graph_order_radius,
    get_distance,
    graph_field_network,
    is_local_edge,
)


class MoleculeGNN(ModelMixin, ConfigMixin):
    """Graph Neural Network Model for molecule conformation.

    Code adapted from colab:
        https://colab.research.google.com/drive/1pLYYWQhdLuv1q-JtEHGZybxp2RBF8gPs#scrollTo=-3-P4w5sXkRU written by Nathan Lambert.
    """

    def __init__(
        self,
        hidden_dim: int = 128,
        num_convs: int = 6,
        num_convs_local: int = 4,
        cutoff: float = 10.0,
        mlp_act: str = "relu",
        edge_order: int = 3,
        edge_encoder: str = "mlp",
        smooth_conv: bool = True,
    ) -> None:
        """Constructor for MoleculeGNN.
        Args:
            hidden_dim: hidden dimension of the GNN.
            num_convs: number of GNN layers.
            num_convs_local: number of GNN layers for local edges.
            cutoff: nutoff radius for the GNN.
            mlp_act: activation function for the MLP.
            edge_order: order of the edge features.
            edge_encoder: type of edge encoder.
            smooth_conv: whether to use smooth convolution.
        """
        super().__init__()
        self.cutoff = cutoff
        self.edge_encoder = edge_encoder
        self.edge_order = edge_order

        # edge_encoder: Takes both edge type and edge length as input and outputs a vector [Note]: node embedding is done in SchNetEncoder
        self.edge_encoder_global = MLPEdgeEncoder(
            hidden_dim, mlp_act
        )  # get_edge_encoder(config)
        self.edge_encoder_local = MLPEdgeEncoder(
            hidden_dim, mlp_act
        )  # get_edge_encoder(config)

        # the graph neural network that extracts node-wise features.
        self.encoder_global = SchNetEncoder(
            hidden_channels=hidden_dim,
            num_filters=hidden_dim,
            num_interactions=num_convs,
            edge_channels=self.edge_encoder_global.out_channels,
            cutoff=cutoff,
            smooth=smooth_conv,
        )
        self.encoder_local = GINEncoder(
            hidden_dim=hidden_dim,
            num_convs=num_convs_local,
        )

        # `output_mlp` takes a mixture of two nodewise features and edge features as input and outputs
        # gradients w.r.t. edge_length (out_dim = 1).
        self.grad_global_dist_mlp = MultiLayerPerceptron(
            2 * hidden_dim, [hidden_dim, hidden_dim // 2, 1], activation=mlp_act
        )

        self.grad_local_dist_mlp = MultiLayerPerceptron(
            2 * hidden_dim, [hidden_dim, hidden_dim // 2, 1], activation=mlp_act
        )

        # incorporate parameters together
        self.model_global = nn.ModuleList(
            [self.edge_encoder_global, self.encoder_global, self.grad_global_dist_mlp]
        )
        self.model_local = nn.ModuleList(
            [self.edge_encoder_local, self.encoder_local, self.grad_local_dist_mlp]
        )

    def _forward(
        self,
        atom_type: torch.Tensor,
        pos: torch.Tensor,
        bond_index: torch.Tensor,
        bond_type: torch.Tensor,
        batch: torch.Tensor,
        edge_index: Optional[torch.Tensor] = None,
        edge_type: Optional[torch.Tensor] = None,
        edge_length: Optional[int] = None,
        return_edges: bool = False,
        extend_order: bool = True,
        extend_radius: bool = True,
        is_sidechain: Optional[bool] = None,
    ) -> Tuple[Any, ...]:
        """Forward pass for edges features.

        Args:
            atom_type:  types of atoms, (N, ).
            pos:        positions of atoms, (N, 3).
            bond_index: indices of bonds (not extended, not radius-graph), (2, E).
            bond_type:  bond types, (E, ).
            batch:      node index to graph index, (N, ).
            edge_index: indices of edges (extended, radius-graph), (2, E'). If None, determined by extend_graph_order_radius.
            edge_type: edge types, (E', ). If None, determined by extend_graph_order_radius.
            edge_length: edge lengths, (E', ). If None, determined by extend_graph_order_radius.
            return_edges: whether to return edge_index, edge_type, edge_length.
            extend_order: whether to extend the graph by bond order.
            extend_radius: whether to extend the graph by radius.
            is_sidechain: whether the atom is a sidechain atom.

        Returns:
            tuple with local and global invariant edge features.
                If return_edges is True, it also returns edge_index, edge_type, edge_length, local_edge_index.

        """
        n = atom_type.size(0)
        if edge_index is None or edge_type is None or edge_length is None:
            edge_index, edge_type = extend_graph_order_radius(
                num_nodes=n,
                pos=pos,
                edge_index=bond_index,
                edge_type=bond_type,
                batch=batch,
                order=self.edge_order,
                cutoff=self.cutoff,
                extend_order=extend_order,
                extend_radius=extend_radius,
                is_sidechain=is_sidechain,
            )
            edge_length = get_distance(pos, edge_index).unsqueeze(-1)  # (E, 1)
        local_edge_mask = is_local_edge(edge_type)  # (E, )

        # with the parameterization of NCSNv2
        # DDPM loss implicit handle the noise variance scale conditioning
        sigma_edge = torch.ones(
            size=(edge_index.size(1), 1), device=pos.device  # type: ignore
        )  # (E, 1)

        # Encoding global
        edge_attr_global = self.edge_encoder_global(
            edge_length=edge_length, edge_type=edge_type
        )  # Embed edges

        # Global
        node_attr_global = self.encoder_global(
            z=atom_type,
            edge_index=edge_index,
            edge_length=edge_length,
            edge_attr=edge_attr_global,
        )
        # Assemble pairwise features
        h_pair_global = assemble_atom_pair_feature(
            node_attr=node_attr_global,
            edge_index=edge_index,
            edge_attr=edge_attr_global,
        )  # (E_global, 2H)
        # Invariant features of edges (radius graph, global)
        edge_inv_global = self.grad_global_dist_mlp(h_pair_global) * (
            1.0 / sigma_edge
        )  # (E_global, 1)

        # Encoding local
        edge_attr_local = self.edge_encoder_global(
            edge_length=edge_length, edge_type=edge_type
        )  # Embed edges
        # edge_attr += temb_edge

        # Local
        node_attr_local = self.encoder_local(
            z=atom_type,
            edge_index=edge_index[:, local_edge_mask],  # type: ignore
            edge_attr=edge_attr_local[local_edge_mask],
        )
        # Assemble pairwise features
        h_pair_local = assemble_atom_pair_feature(
            node_attr=node_attr_local,
            edge_index=edge_index[:, local_edge_mask],  # type: ignore
            edge_attr=edge_attr_local[local_edge_mask],
        )  # (E_local, 2H)

        # Invariant features of edges (bond graph, local)
        if isinstance(sigma_edge, torch.Tensor):
            edge_inv_local = self.grad_local_dist_mlp(h_pair_local) * (
                1.0 / sigma_edge[local_edge_mask]
            )  # (E_local, 1)
        else:
            edge_inv_local = self.grad_local_dist_mlp(h_pair_local) * (
                1.0 / sigma_edge
            )  # (E_local, 1)

        if return_edges:
            return (
                edge_inv_global,
                edge_inv_local,
                edge_index,
                edge_type,
                edge_length,
                local_edge_mask,
            )
        else:
            return (edge_inv_global, edge_inv_local)

    def forward(
        self,
        sample: Data,
        timestep: Union[float, int],
        return_dict: bool = True,
        sigma: float = 1.0,
        global_start_sigma: float = 0.5,
        w_global: float = 1.0,
        extend_order=False,
        extend_radius=True,
        clip_local=None,
        clip_global: float = 1000.0,
    ) -> Union[MoleculeGNNOutput, Tuple]:
        """Forward pass for the model.

        Args:
            sample: packed torch geometric object.
            timestep timesteps.
            return_dict: whether or not to return a MoleculeGNNOutput instead of a plain tuple.
            sigma: the noise variance scale.
            global_start_sigma: the noise variance scale for global edges.
            w_global: the weight for global edges.
            extend_order: whether to extend the graph by bond order.
            extend_radius: whether to extend the graph by radius.
            clip_local: the clip value for local edges.
            clip_global: the clip value for global edges.

        Returns:
            dict with samples. If return_dict is false, return positions.
        """

        # unpack sample
        atom_type = sample.atom_type
        bond_index = sample.edge_index
        bond_type = sample.edge_type
        pos = sample.pos

        (
            edge_inv_global,
            edge_inv_local,
            edge_index,
            _,
            edge_length,
            local_edge_mask,
        ) = self._forward(
            atom_type=atom_type,
            pos=sample.pos,
            bond_index=bond_index,
            bond_type=bond_type,
            batch=sample.batch,
            return_edges=True,
            extend_order=extend_order,
            extend_radius=extend_radius,
        )  # (E_global, 1), (E_local, 1)

        # Important equation in the paper for equivariant features - eqns 5-7 of GeoDiff
        node_eq_local = graph_field_network(
            edge_inv_local,
            pos,
            edge_index[:, local_edge_mask],
            edge_length[local_edge_mask],
        )
        if clip_local is not None:
            node_eq_local = clip_norm(node_eq_local, limit=clip_local)

        # Global
        if sigma < global_start_sigma:
            edge_inv_global = edge_inv_global * (
                1 - local_edge_mask.view(-1, 1).float()
            )
            node_eq_global = graph_field_network(
                edge_inv_global, pos, edge_index, edge_length
            )
            node_eq_global = clip_norm(node_eq_global, limit=clip_global)
        else:
            node_eq_global = 0

        # Sum
        eps_pos = node_eq_local + node_eq_global * w_global

        if not return_dict:
            return (-eps_pos,)

        return MoleculeGNNOutput(sample=torch.FloatTensor(-eps_pos).to(pos.device))  # type: ignore



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""MolGX Algorithm.

MolGX generation algorithm.
"""

import logging
from dataclasses import field
from typing import Any, ClassVar, Dict, Iterator, Optional, TypeVar

from ....domains.materials import SMILES, MoleculeFormat, validate_molecules
from ....exceptions import InvalidItem
from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
from ...registry import ApplicationsRegistry
from .implementation import MolGXGenerator

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = type(None)
S = TypeVar("S", bound=SMILES)


class MolGX(GeneratorAlgorithm[S, T]):
    """MolGX Algorithm."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T] = None,
    ):
        """Instantiate MolGX ready to generate items.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for generating small molecules (SMILES) with given HOMO and LUMO energies:

                configuration = MolGXQM9Generator()
                molgx = MolGX(configuration=configuration, target=target)
                items = list(molgx.sample(10))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input
        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to sample batches via the ConditionalGenerator.

        Args:
                configuration: helps to set up the application.
                target: context or condition for the generation. Unused in the algorithm.

        Returns:
                callable generating a batch of items.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: MolGXGenerator = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.generate

    def validate_configuration(
        self, configuration: AlgorithmConfiguration[S, T]
    ) -> AlgorithmConfiguration[S, T]:
        # TODO raise InvalidAlgorithmConfiguration
        assert isinstance(configuration, AlgorithmConfiguration)
        return configuration

    def sample(self, number_of_items: int = 100) -> Iterator[S]:
        """Generate a number of unique and valid items.

        Args:
            number_of_items: number of items to generate. Defaults to 100.

        Yields:
            the items.
        """
        if hasattr(self.configuration, "maximum_number_of_solutions"):
            maxiumum_number_of_molecules = int(
                getattr(self.configuration, "maximum_number_of_solutions")
            )
            if number_of_items > maxiumum_number_of_molecules:
                logger.warning(
                    f"current MolGX configuration can not support generation of {number_of_items} molecules..."
                )
                logger.warning(
                    f"to enable generation of {number_of_items} molecules, increase 'maximum_number_of_solutions' (currently set to {maxiumum_number_of_molecules})"
                )
                number_of_items = maxiumum_number_of_molecules
                logger.warning(f"generating at most: {maxiumum_number_of_molecules}...")

        return super().sample(number_of_items=number_of_items)


@ApplicationsRegistry.register_algorithm_application(MolGX)
class MolGXQM9Generator(AlgorithmConfiguration[SMILES, Any]):
    """Configuration to generate compounds with given HOMO and LUMO energies."""

    algorithm_type: ClassVar[str] = "conditional_generation"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"
    homo_energy_value: float = field(
        default=-0.25,
        metadata=dict(description="Target HOMO energy value."),
    )
    lumo_energy_value: float = field(
        default=0.08,
        metadata=dict(description="Target LUMO energy value."),
    )
    use_linear_model: bool = field(
        default=True,
        metadata=dict(description="Linear model usage."),
    )
    number_of_candidates: int = field(
        default=2,
        metadata=dict(description="Number of candidates to consider."),
    )
    maximum_number_of_candidates: int = field(
        default=5,
        metadata=dict(description="Maximum number of candidates to consider."),
    )
    maximum_number_of_solutions: int = field(
        default=10,
        metadata=dict(description="Maximum number of solutions."),
    )
    maximum_number_of_nodes: int = field(
        default=50000,
        metadata=dict(description="Maximum number of nodes in the graph exploration."),
    )
    beam_size: int = field(
        default=2000,
        metadata=dict(description="Size of the beam during search."),
    )
    without_estimate: bool = field(
        default=True,
        metadata=dict(description="Disable estimates."),
    )
    use_specific_rings: bool = field(
        default=True,
        metadata=dict(description="Flag to indicate whether specific rings are used."),
    )
    use_fragment_const: bool = field(
        default=False,
        metadata=dict(description="Using constant fragments."),
    )

    def get_target_description(self) -> Optional[Dict[str, str]]:
        """Get description of the target for generation.

        Returns:
                target description, returns None in case no target is used.
        """
        return None

    def get_conditional_generator(self, resources_path: str) -> MolGXGenerator:
        """Instantiate the actual generator implementation.

        Args:
                resources_path: local path to model files.

        Returns:
                instance with :meth:`generate<gt4sd.algorithms.conditional_generation.molgx.implementation.MolGXGenerator.generate>` for generation.
        """
        return MolGXGenerator(
            resources_path=resources_path,
            homo_energy_value=self.homo_energy_value,
            lumo_energy_value=self.lumo_energy_value,
            use_linear_model=self.use_linear_model,
            number_of_candidates=self.number_of_candidates,
            maximum_number_of_candidates=self.maximum_number_of_candidates,
            maximum_number_of_solutions=self.maximum_number_of_solutions,
            maximum_number_of_nodes=self.maximum_number_of_nodes,
            beam_size=self.beam_size,
            without_estimate=self.without_estimate,
            use_specific_rings=self.use_specific_rings,
            use_fragment_const=self.use_fragment_const,
            tag_name="qm9_sample_pretrained_model.pickle",
        )

    def validate_item(self, item: str) -> SMILES:
        """Check that item is a valid SMILES.

        Args:
                item: a generated item that is possibly not valid.

        Raises:
                InvalidItem: in case the item can not be validated.

        Returns:
                the validated SMILES.
        """
        (
            molecules,
            _,
        ) = validate_molecules([item], MoleculeFormat.smiles)
        if molecules[0] is None:
            raise InvalidItem(
                title="InvalidSMILES",
                detail=f'rdkit.Chem.MolFromSmiles returned None for "{item}"',
            )
        return SMILES(item)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Prediction algorithms based on PaccMann"""

import logging
from dataclasses import field
from typing import Any, ClassVar, List, Optional, TypeVar

from ...core import AlgorithmConfiguration, GeneratorAlgorithm, Untargeted
from ...registry import ApplicationsRegistry
from .implementation import BimodalMCAAffinityPredictor, MCAPredictor

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

T = TypeVar("T", bound=Any)
S = TypeVar("S", bound=Any)


class PaccMann(GeneratorAlgorithm[S, T]):
    """PaccMann predictor."""

    def __init__(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T] = None,
    ):
        """Instantiate PaccMann for prediction.

        Args:
            configuration: domain and application
                specification defining parameters, types and validations.
            target: a target for which to generate items.

        Example:
            An example for predicting affinity for a given ligand and target protein pair::

                config = AffinityPredictor(
                    protein_targets=[
                        "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
                        "MVLSPADKTNVKAAWGKVGAHAGEYGAEALERMFLSFPTT",
                    ],
                    ligands=[
                        "CONN=COc1cc2ccccc2c11Occncc(Cl)c1N(O)O",
                        "ClCCC(O1)C(C(N=C1C(=O)Nc1cccc(F)c1F)SO)C",
                    ]
                )
                algorithm = PaccMann(configuration=config)
                items = list(algorithm.sample(1))
                print(items)
        """

        configuration = self.validate_configuration(configuration)
        # TODO there might also be a validation/check on the target input

        super().__init__(
            configuration=configuration,  # type:ignore
            target=target,  # type:ignore
        )

    def get_generator(
        self,
        configuration: AlgorithmConfiguration[S, T],
        target: Optional[T],
    ) -> Untargeted:
        """Get the function to perform the prediction via PaccMann's generator.

        Args:
            configuration: helps to set up specific application of PaccMann.
            target: context or condition for the generation.

        Returns:
            callable with target predicting properties using PaccMann.
        """
        logger.info("ensure artifacts for the application are present.")
        self.local_artifacts = configuration.ensure_artifacts()
        implementation: MCAPredictor = configuration.get_conditional_generator(  # type: ignore
            self.local_artifacts
        )
        return implementation.predict_values


@ApplicationsRegistry.register_algorithm_application(PaccMann)
class AffinityPredictor(AlgorithmConfiguration[str, str]):
    """Configuration to predict affinity for a given ligand/protrin target pair."""

    algorithm_type: ClassVar[str] = "prediction"
    domain: ClassVar[str] = "materials"
    algorithm_version: str = "v0"

    protein_targets: List[str] = field(
        default_factory=list,
        metadata=dict(description="List of protein targets as AA sequences."),
    )
    ligands: List[str] = field(
        default_factory=list,
        metadata=dict(description="List of ligands in SMILES format."),
    )
    confidence: bool = field(
        default=False,
        metadata=dict(
            description="Whether the confidence for the prediction should be returned."
        ),
    )

    def get_conditional_generator(
        self, resources_path: str
    ) -> BimodalMCAAffinityPredictor:
        """Instantiate the actual predictor implementation.

        Args:
            resources_path: local path to model files.

        Returns:
            instance with :meth:`gt4sd.algorithms.prediction.affinity._predicto.implementation.BimodalMCAAffinityPredictor.predict` method for predicting affinity.
        """
        return BimodalMCAAffinityPredictor(
            resources_path=resources_path,
            protein_targets=self.protein_targets,
            ligands=self.ligands,
            confidence=self.confidence,
        )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation of MolGX conditional generators."""

import logging
import os
from typing import Any, Dict, List

from molgx.molgxsdk import MolgxSdk

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class MolGXGenerator:
    """Interface for MolGX generator."""

    def __init__(
        self,
        resources_path: str,
        tag_name: str,
        homo_energy_value: float = -0.25,
        lumo_energy_value: float = 0.08,
        use_linear_model: bool = True,
        number_of_candidates: int = 2,
        maximum_number_of_candidates: int = 3,
        maximum_number_of_solutions: int = 3,
        maximum_number_of_nodes: int = 50000,
        beam_size: int = 2000,
        without_estimate: bool = True,
        use_specific_rings: bool = True,
        use_fragment_const: bool = False,
    ) -> None:
        """Instantiate a MolGX generator.

        Args:
                resources_path: path to the resources for model loading.
                tag_name: tag for the pretrained model.
                homo_energy_value: target HOMO energy value. Defaults to -0.25.
                lumo_energy_value: target LUMO energy value. Defaults to 0.08.
                use_linear_model: linear model usage. Defaults to True.
                number_of_candidates: number of candidates to consider. Defaults to 2.
                maximum_number_of_candidates: maximum number of candidates to consider. Defaults to 3.
                maximum_number_of_solutions: maximum number of solutions. Defaults to 3.
                maximum_number_of_nodes: maximum number of nodes in the graph exploration. Defaults to 50000.
                beam_size: size of the beam during search. Defaults to 2000.
                without_estimate: disable estimates. Defaults to True.
                use_specific_rings: flag to indicate whether specific rings are used. Defaults to True.
                use_fragment_const: using constant fragments. Defaults to False.

        Raises:
                RuntimeError: in the case extras are disabled.
        """

        # loading artifacts
        self.resources_path = resources_path
        self.tag_name = tag_name
        self.amd = self.load_molgx(self.resources_path, self.tag_name)
        self.molecules_data, self.target_property = self.amd.LoadPickle()
        # algorithm parameters
        self._homo_energy_value = homo_energy_value
        self._lumo_energy_value = lumo_energy_value
        self._use_linear_model = use_linear_model
        self._number_of_candidates = number_of_candidates
        self._maximum_number_of_candidates = maximum_number_of_candidates
        self._maximum_number_of_solutions = maximum_number_of_solutions
        self._maximum_number_of_nodes = maximum_number_of_nodes
        self._beam_size = beam_size
        self._without_estimate = without_estimate
        self._use_specific_rings = use_specific_rings
        self._use_fragment_const = use_fragment_const
        self._parameters = self._create_parameters_dictionary()

    @staticmethod
    def load_molgx(resource_path: str, tag_name: str) -> MolgxSdk:
        """Load MolGX model.

        Args:
                resource_path: path to the resources for model loading.
                tag_name: tag for the pretrained model.

        Returns:
                MolGX model SDK.
        """
        return MolgxSdk(
            pickle_filepath=os.path.join(resource_path, tag_name),
        )

    def _create_parameters_dictionary(self) -> Dict[str, Any]:
        """Create parameters dictionary.

        Returns:
                the parameters to run MolGX.
        """
        self.target_property["homo"] = (-10, 10)
        self.target_property["lumo"] = (-10, 10)
        parameters: Dict[str, Any] = {}
        parameters["target_property"] = self.target_property
        parameters["use_linear_model"] = self.use_linear_model
        parameters["num_candidate"] = self.number_of_candidates
        parameters["max_candidate"] = self.maximum_number_of_candidates
        parameters["max_solution"] = self.maximum_number_of_solutions
        parameters["max_node"] = self.maximum_number_of_nodes
        parameters["beam_size"] = self.beam_size
        parameters["without_estimate"] = self.without_estimate
        parameters["use_specific_rings"] = self.use_specific_rings
        parameters["use_fragment_const"] = self.use_fragment_const
        return parameters

    @property
    def homo_energy_value(self) -> float:
        return self._homo_energy_value

    @homo_energy_value.setter
    def homo_energy_value(self, value: float) -> None:
        self._homo_energy_value = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def lumo_energy_value(self) -> float:
        return self._lumo_energy_value

    @lumo_energy_value.setter
    def lumo_energy_value(self, value: float) -> None:
        self._lumo_energy_value = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def use_linear_model(self) -> bool:
        return self._use_linear_model

    @use_linear_model.setter
    def use_linear_model(self, value: bool) -> None:
        self._use_linear_model = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def number_of_candidates(self) -> int:
        return self._number_of_candidates

    @number_of_candidates.setter
    def number_of_candidates(self, value: int) -> None:
        self._number_of_candidates = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def maximum_number_of_candidates(self) -> int:
        return self._maximum_number_of_candidates

    @maximum_number_of_candidates.setter
    def maximum_number_of_candidates(self, value: int) -> None:
        self._maximum_number_of_candidates = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def maximum_number_of_solutions(self) -> int:
        return self._maximum_number_of_solutions

    @maximum_number_of_solutions.setter
    def maximum_number_of_solutions(self, value: int) -> None:
        self._maximum_number_of_solutions = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def maximum_number_of_nodes(self) -> int:
        return self._maximum_number_of_nodes

    @maximum_number_of_nodes.setter
    def maximum_number_of_nodes(self, value: int) -> None:
        self._maximum_number_of_nodes = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def beam_size(self) -> int:
        return self._beam_size

    @beam_size.setter
    def beam_size(self, value: int) -> None:
        self._beam_size = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def without_estimate(self) -> bool:
        return self._without_estimate

    @without_estimate.setter
    def without_estimate(self, value: bool) -> None:
        self._without_estimate = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def use_specific_rings(self) -> bool:
        return self._use_specific_rings

    @use_specific_rings.setter
    def use_specific_rings(self, value: bool) -> None:
        self._use_specific_rings = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def use_fragment_const(self) -> bool:
        return self._use_fragment_const

    @use_fragment_const.setter
    def use_fragment_const(self, value: bool) -> None:
        self._use_fragment_const = value
        self.parameters = self._create_parameters_dictionary()

    @property
    def parameters(self) -> Dict[str, Any]:
        return self._parameters

    @parameters.setter
    def parameters(self, value: Dict[str, Any]) -> None:
        parameters = self._create_parameters_dictionary()
        parameters.update(value)
        self._parameters = parameters

    def generate(self) -> List[str]:
        """Sample random molecules.

        Returns:
                sampled molecule (SMILES).
        """
        # generate molecules
        logger.info(f"running MolGX with the following parameters: {self.parameters}")
        print(self.parameters)
        molecules_df = self.amd.GenMols(self.molecules_data, self.parameters)
        logger.info("MolGX run completed")
        return molecules_df["SMILES"].tolist()



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Data module."""

from __future__ import division, print_function

import csv
import functools
import json
import logging
import os
import random
from typing import Any, Callable, List, Tuple, Union, Optional

import numpy as np
import torch
from pymatgen.core.structure import Structure  # type: ignore
from torch import LongTensor, Tensor
from torch.utils.data import DataLoader, Dataset
from torch.utils.data.dataloader import default_collate
from torch.utils.data.sampler import SubsetRandomSampler

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def get_train_val_test_loader(
    dataset: torch.utils.data.Dataset,
    collate_fn: Callable[[List[Any]], Any] = default_collate,
    batch_size: int = 64,
    train_ratio: Optional[float] = None,
    val_ratio: float = 0.1,
    test_ratio: float = 0.1,
    return_test: bool = False,
    num_workers: int = 1,
    pin_memory: bool = False,
    **kwargs,
) -> Union[
    Tuple[DataLoader[Any], DataLoader[Any], DataLoader[Any]],
    Tuple[DataLoader[Any], DataLoader[Any]],
]:
    """Utility function for dividing a dataset to train, val, test datasets.

    !!! The dataset needs to be shuffled before using the function !!!

    Args:
        dataset: torch.utils.data.Dataset
          The full dataset to be divided.
        collate_fn: torch.utils.data.DataLoader.
        batch_size: int.
        train_ratio: float.
        val_ratio: float.
        test_ratio: float.
        return_test: bool.
          Whether to return the test dataset loader. If False, the last test_size
          data will be hidden.
        num_workers: int.
        pin_memory: bool.

    Returns:
        train_loader: torch.utils.data.DataLoader
          DataLoader that random samples the training data.
        val_loader: torch.utils.data.DataLoader
          DataLoader that random samples the validation data.
        (test_loader): torch.utils.data.DataLoader
          DataLoader that random samples the test data, Returns if
            return_test=True.
    """
    total_size = len(dataset)  # type: ignore
    if kwargs["train_size"] is None:
        if train_ratio is None:
            assert val_ratio + test_ratio < 1
            train_ratio = 1 - val_ratio - test_ratio
            logger.warning(
                f"train_ratio is None, using 1 - val_ratio - "
                f"test_ratio = {train_ratio} as training data."
            )

        else:
            assert train_ratio + val_ratio + test_ratio <= 1
    indices = list(range(total_size))
    if kwargs["train_size"]:
        train_size = kwargs["train_size"]
    else:
        train_size = int(train_ratio * total_size)  # type: ignore
    if kwargs["test_size"]:
        test_size = kwargs["test_size"]
    else:
        test_size = int(test_ratio * total_size)
    if kwargs["val_size"]:
        valid_size = kwargs["val_size"]
    else:
        valid_size = int(val_ratio * total_size)
    train_sampler = SubsetRandomSampler(indices[:train_size])
    val_sampler = SubsetRandomSampler(indices[-(valid_size + test_size) : -test_size])

    train_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=train_sampler,
        num_workers=num_workers,
        collate_fn=collate_fn,
        pin_memory=pin_memory,
    )
    val_loader = DataLoader(
        dataset,
        batch_size=batch_size,
        sampler=val_sampler,
        num_workers=num_workers,
        collate_fn=collate_fn,
        pin_memory=pin_memory,
    )
    if return_test:

        test_sampler = SubsetRandomSampler(indices[-test_size:])

        test_loader = DataLoader(
            dataset,
            batch_size=batch_size,
            sampler=test_sampler,
            num_workers=num_workers,
            collate_fn=collate_fn,
            pin_memory=pin_memory,
        )

        return train_loader, val_loader, test_loader
    else:
        return train_loader, val_loader


def collate_pool(
    dataset_list: List[Any],
) -> Tuple[Tuple[Tensor, Tensor, Tensor, List[LongTensor]], Tensor, List[Any]]:
    """Collate a list of data and return a batch for predicting crystal properties.

    Args:
        dataset_list: list of tuples for each data point.
          (atom_fea, nbr_fea, nbr_fea_idx, target)

          atom_fea: torch.Tensor shape (n_i, atom_fea_len).
          nbr_fea: torch.Tensor shape (n_i, M, nbr_fea_len).
          nbr_fea_idx: torch.LongTensor shape (n_i, M).
          target: torch.Tensor shape (1, ).
          cif_id: str or int.

    Returns:
        N = sum(n_i); N0 = sum(i)
        batch_atom_fea: torch.Tensor shape (N, orig_atom_fea_len)
          Atom features from atom type.
        batch_nbr_fea: torch.Tensor shape (N, M, nbr_fea_len)
          Bond features of each atom's M neighbors.
        batch_nbr_fea_idx: torch.LongTensor shape (N, M)
          Indices of M neighbors of each atom.
        crystal_atom_idx: list of torch.LongTensor of length N0
          Mapping from the crystal idx to atom idx.
        target: torch.Tensor shape (N, 1)
          Target value for prediction.
        batch_cif_ids: list.
    """
    batch_atom_fea, batch_nbr_fea, batch_nbr_fea_idx = [], [], []
    crystal_atom_idx, batch_target = [], []
    batch_cif_ids = []
    base_idx = 0
    for i, ((atom_fea, nbr_fea, nbr_fea_idx), target, cif_id) in enumerate(
        dataset_list
    ):
        n_i = atom_fea.shape[0]  # number of atoms for this crystal
        batch_atom_fea.append(atom_fea)
        batch_nbr_fea.append(nbr_fea)
        batch_nbr_fea_idx.append(nbr_fea_idx + base_idx)
        new_idx = torch.LongTensor(np.arange(n_i) + base_idx)
        crystal_atom_idx.append(new_idx)
        batch_target.append(target)
        batch_cif_ids.append(cif_id)
        base_idx += n_i
    return (
        (
            torch.cat(batch_atom_fea, dim=0),
            torch.cat(batch_nbr_fea, dim=0),
            torch.cat(batch_nbr_fea_idx, dim=0),
            crystal_atom_idx,
        ),
        torch.stack(batch_target, dim=0),
        batch_cif_ids,
    )


class GaussianDistance:
    """Expands the distance by Gaussian basis.

    Unit: angstrom
    """

    def __init__(
        self, dmin: float, dmax: float, step: float, var: Optional[float] = None
    ):
        """
        Args:
            dmin: float
              Minimum interatomic distance.
            dmax: float
              Maximum interatomic distance.
            step: float
              Step size for the Gaussian filter.
        """
        assert dmin < dmax
        assert dmax - dmin > step
        self.filter = np.arange(dmin, dmax + step, step)
        if var is None:
            var = step
        self.var = var

    def expand(self, distances: np.ndarray) -> np.ndarray:
        """Apply Gaussian disntance filter to a numpy distance array.

        Args:
            distance: np.array shape n-d array
              A distance matrix of any shape.

        Returns:
            expanded_distance: shape (n+1)-d array
              Expanded distance matrix with the last dimension of length
              len(self.filter).
        """
        return np.exp(
            -((distances[..., np.newaxis] - self.filter) ** 2) / self.var**2
        )


class AtomInitializer:
    """Base class for intializing the vector representation for atoms.

    !!! Use one AtomInitializer per dataset !!!
    """

    def __init__(self, atom_types):
        self.atom_types = set(atom_types)
        self._embedding = {}

    def get_atom_fea(self, atom_type):
        assert atom_type in self.atom_types
        return self._embedding[atom_type]

    def load_state_dict(self, state_dict):
        self._embedding = state_dict
        self.atom_types = set(self._embedding.keys())
        self._decodedict = {
            idx: atom_type for atom_type, idx in self._embedding.items()
        }

    def state_dict(self):
        return self._embedding

    def decode(self, idx):
        if not hasattr(self, "_decodedict"):
            self._decodedict = {
                idx: atom_type for atom_type, idx in self._embedding.items()
            }
        return self._decodedict[idx]


class AtomCustomJSONInitializer(AtomInitializer):
    """
    Initialize atom feature vectors using a JSON file, which is a python
    dictionary mapping from element number to a list representing the
    feature vector of the element.

    """

    def __init__(self, elem_embedding_file: str):
        """
        Args:
            elem_embedding_file: str
                The path to the .json file.
        """
        with open(elem_embedding_file) as f:
            elem_embedding = json.load(f)
        elem_embedding = {int(key): value for key, value in elem_embedding.items()}
        atom_types = set(elem_embedding.keys())
        super(AtomCustomJSONInitializer, self).__init__(atom_types)
        for key, value in elem_embedding.items():
            self._embedding[key] = np.array(value, dtype=float)


class CIFData(Dataset):
    """
    The CIFData dataset is a wrapper for a dataset where the crystal structures
    are stored in the form of CIF files. The dataset should have the following
    directory structure:

    root_dir
    ├── id_prop.csv
    ├── atom_init.json
    ├── id0.cif
    ├── id1.cif
    ├── ...

    id_prop.csv: a CSV file with two columns. The first column recodes a
    unique ID for each crystal, and the second column recodes the value of
    target property.

    atom_init.json: a JSON file that stores the initialization vector for each
    element.

    ID.cif: a CIF file that recodes the crystal structure, where ID is the
    unique ID for the crystal.
    """

    def __init__(
        self,
        root_dir: str,
        max_num_nbr: int = 12,
        radius: int = 8,
        dmin: int = 0,
        step: float = 0.2,
        random_seed: int = 123,
        atom_initialization: Optional[AtomCustomJSONInitializer] = None,
    ):
        """
        Args:
            root_dir: str
                The path to the root directory of the dataset.
            max_num_nbr: int
                The maximum number of neighbors while constructing the crystal graph.
            radius: float
                The cutoff radius for searching neighbors.
            dmin: float
                The minimum distance for constructing GaussianDistance.
            step: float
                The step size for constructing GaussianDistance.
            random_seed: int
                Random seed for shuffling the dataset.
            atom_initialization: AtomInitializer
                The atom initializer for initializing the atom feature vectors.
                Defaults to None, in which case a `atom_init.json` should be in `root_dir`.
        """
        self.root_dir = root_dir
        self.max_num_nbr, self.radius = max_num_nbr, radius
        assert os.path.exists(root_dir), "root_dir does not exist!"

        id_prop_file = os.path.join(self.root_dir, "id_prop.csv")

        self.training = os.path.exists(id_prop_file)
        if self.training:
            with open(id_prop_file) as f:
                reader = csv.reader(f)
                self.id_prop_data = [row for row in reader]
        else:
            # Emulate the label file
            self.id_prop_data = [  # type: ignore
                (os.path.splitext(cif_id)[0], "NaN")  # type: ignore
                for cif_id in os.listdir(self.root_dir)
                if cif_id.endswith(".cif")
            ]
        random.shuffle(self.id_prop_data)
        random.seed(random_seed)

        if atom_initialization is None:
            atom_init_file = os.path.join(self.root_dir, "atom_init.json")
            assert os.path.exists(atom_init_file), "atom_init.json does not exist!"
            self.ari = AtomCustomJSONInitializer(atom_init_file)
        else:
            if not isinstance(atom_initialization, AtomInitializer):
                raise TypeError(
                    "atom_initialization should be an instance of AtomInitializer, "
                    f"not {type(atom_initialization)}"
                )
            self.ari = atom_initialization

        self.gdf = GaussianDistance(dmin=dmin, dmax=self.radius, step=step)

    def __len__(self):
        return len(self.id_prop_data)

    @functools.lru_cache(maxsize=None)  # Cache loaded structures
    def __getitem__(self, idx: int) -> Tuple[Any, Any, Any]:  # type: ignore
        """
        Args:
           idx: index.
        Returns:
            atom_fea: torch.Tensor shape (n_i, atom_fea_len).
            nbr_fea: torch.Tensor shape (n_i, M, nbr_fea_len).
            nbr_fea_idx: torch.LongTensor shape (n_i, M).
            target: torch.Tensor shape (1, ).
            cif_id: str or int.
        """
        cif_id, target = self.id_prop_data[idx]
        crystal = Structure.from_file(os.path.join(self.root_dir, cif_id + ".cif"))
        atom_fea = np.vstack(
            [
                self.ari.get_atom_fea(crystal[i].specie.number)
                for i in range(len(crystal))
            ]
        )
        atom_fea = torch.Tensor(atom_fea)  # type: ignore
        all_nbrs = crystal.get_all_neighbors(self.radius, include_index=True)
        all_nbrs = [sorted(nbrs, key=lambda x: x[1]) for nbrs in all_nbrs]
        nbr_fea_idx, nbr_fea = [], []
        for nbr in all_nbrs:
            if len(nbr) < self.max_num_nbr:
                logger.warning(
                    "{} not find enough neighbors to build graph. "
                    "If it happens frequently, consider increase "
                    "radius.".format(cif_id)
                )
                nbr_fea_idx.append(
                    list(map(lambda x: x[2], nbr)) + [0] * (self.max_num_nbr - len(nbr))
                )
                nbr_fea.append(
                    list(map(lambda x: x[1], nbr))
                    + [self.radius + 1.0] * (self.max_num_nbr - len(nbr))
                )
            else:
                nbr_fea_idx.append(list(map(lambda x: x[2], nbr[: self.max_num_nbr])))
                nbr_fea.append(list(map(lambda x: x[1], nbr[: self.max_num_nbr])))
        nbr_fea_idx, nbr_fea = np.array(nbr_fea_idx), np.array(nbr_fea)  # type: ignore
        nbr_fea = self.gdf.expand(nbr_fea)  # type: ignore
        atom_fea = torch.Tensor(atom_fea)  # type: ignore
        nbr_fea = torch.Tensor(nbr_fea)  # type: ignore
        nbr_fea_idx = torch.LongTensor(nbr_fea_idx)  # type: ignore
        target = torch.Tensor([float(target)])  # type: ignore
        return (atom_fea, nbr_fea, nbr_fea_idx), target, cif_id



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Parsing utilities."""

import argparse
import ast
from typing import Union


def str2bool(s: Union[str, bool]) -> bool:
    """Convert a string into a bool.

    Args:
        s: a string representation of a boolean.

    Raises:
        argparse.ArgumentTypeError: in case the conversion is failing.

    Returns:
        the converted value.
    """
    if isinstance(s, bool):
        return s
    if s.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif s.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def convert_string_to_class(s: str):
    """Convert a string into a python object.

    Fallback to ast in case of unexpected strings.

    Args:
        s: a string.

    Returns:
        the converted python object.
    """
    if s.lower() == "true":
        return True
    elif s.lower() == "false":
        return False
    elif s.lower() == "none":
        return None
    elif s:
        try:
            return ast.literal_eval(s)
        except (ValueError, SyntaxError):
            return s



#
# MIT License
#
# Copyright (c) 2024 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import warnings
from gt4sd.frameworks.enzeptional import (
    EnzymeOptimizer,
    SequenceMutator,
    SequenceScorer,
    CrossoverGenerator,
    HuggingFaceEmbedder,
    HuggingFaceModelLoader,
    HuggingFaceTokenizerLoader,
    SelectionGenerator,
)

from gt4sd.configuration import sync_algorithm_with_s3
from gt4sd.configuration import GT4SDConfiguration

configuration = GT4SDConfiguration.get_instance()


warnings.simplefilter(action="ignore", category=FutureWarning)

sync_algorithm_with_s3("proteins/enzeptional/scorers", module="properties")

scorer_filepath = f"{configuration.gt4sd_local_cache_path}/properties/proteins/enzeptional/scorers/feasibility/model.pkl"


def test_optimize():
    language_model_path = "facebook/esm2_t33_650M_UR50D"
    tokenizer_path = "facebook/esm2_t33_650M_UR50D"
    chem_model_path = "seyonec/ChemBERTa-zinc-base-v1"
    chem_tokenizer_path = "seyonec/ChemBERTa-zinc-base-v1"

    model_loader = HuggingFaceModelLoader()
    tokenizer_loader = HuggingFaceTokenizerLoader()

    protein_model = HuggingFaceEmbedder(
        model_loader=model_loader,
        tokenizer_loader=tokenizer_loader,
        model_path=language_model_path,
        tokenizer_path=tokenizer_path,
        cache_dir=None,
        device="cpu",
    )

    chem_model = HuggingFaceEmbedder(
        model_loader=model_loader,
        tokenizer_loader=tokenizer_loader,
        model_path=chem_model_path,
        tokenizer_path=chem_tokenizer_path,
        cache_dir=None,
        device="cpu",
    )

    mutation_config = {
        "type": "language-modeling",
        "embedding_model_path": language_model_path,
        "tokenizer_path": tokenizer_path,
        "unmasking_model_path": language_model_path,
    }

    intervals = [(5, 10), (20, 25)]
    batch_size = 2
    top_k = 1
    substrate_smiles = "NC1=CC=C(N)C=C1"
    product_smiles = "CNC1=CC=C(NC(=O)C2=CC=C(C=C2)C(C)=O)C=C1"

    sample_sequence = "MSKLLMIGTGPVAIDQFLTRYEASCQAYKDMHQDQQLSSQFNTNLFEGDKALVTKFLEINRTLS"

    mutator = SequenceMutator(sequence=sample_sequence, mutation_config=mutation_config)
    mutator.set_top_k(top_k)

    scorer = SequenceScorer(
        protein_model=protein_model,
        scorer_filepath=scorer_filepath,
        use_xgboost=False,
        scaler_filepath=None,
    )

    selection_generator = SelectionGenerator()
    crossover_generator = CrossoverGenerator()

    optimizer = EnzymeOptimizer(
        sequence=sample_sequence,
        mutator=mutator,
        scorer=scorer,
        intervals=intervals,
        substrate_smiles=substrate_smiles,
        product_smiles=product_smiles,
        chem_model=chem_model,
        selection_generator=selection_generator,
        crossover_generator=crossover_generator,
        concat_order=["substrate", "sequence", "product"],
        batch_size=batch_size,
        selection_ratio=0.25,
        perform_crossover=True,
        crossover_type="single_point",
        pad_intervals=False,
        minimum_interval_length=8,
        seed=123,
    )

    num_iterations = 3
    num_sequences = 5
    num_mutations = 5
    time_budget = 50000

    optimized_sequences, iteration_info = optimizer.optimize(
        num_iterations=num_iterations,
        num_sequences=num_sequences,
        num_mutations=num_mutations,
        time_budget=time_budget,
    )

    assert len(optimized_sequences) > 0



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Implementation of scorers."""
from functools import partial
from typing import Any, Callable, Dict, List, Type, Optional

import numpy as np
from rdkit import Chem
from guacamol.common_scoring_functions import (
    IsomerScoringFunction,
    RdkitScoringFunction,
    SMARTSScoringFunction,
    TanimotoScoringFunction,
)
from guacamol.score_modifier import (
    ClippedScoreModifier,
    GaussianModifier,
    MaxGaussianModifier,
    MinGaussianModifier,
)
from guacamol.scoring_function import ScoringFunction
from guacamol.utils.descriptors import (
    bertz,
    logP,
    mol_weight,
    num_aromatic_rings,
    num_rings,
    num_rotatable_bonds,
    qed,
    tpsa,
)

MODIFIERS: Dict[str, Callable[..., Any]] = {
    "gaussian_modifier": GaussianModifier,
    "min_gaussian_modifier": MinGaussianModifier,
    "max_gaussian_modifier": MaxGaussianModifier,
    "clipped_score_modifier": ClippedScoreModifier,
}
MODIFIERS_PARAMETERS: Dict[str, Dict[str, float]] = {
    "gaussian_modifier": {"mu": 2, "sigma": 0.5},
    "min_gaussian_modifier": {"mu": 0.75, "sigma": 0.1},
    "max_gaussian_modifier": {"mu": 100, "sigma": 10},
    "clipped_score_modifier": {"upper_x": 0.8},
}
DESCRIPTOR: Dict[str, Callable[..., Any]] = {
    "num_rotatable_bonds": num_rotatable_bonds,
    "num_aromatic_rings": num_aromatic_rings,
    "log_p": logP,
    "tpsa": tpsa,
    "bertz": bertz,
    "qed": qed,
    "mol_weight": mol_weight,
    "num_rings": num_rings,
}


def distance_to_score(distance: float, beta: float) -> float:
    """calculating exponential for a given distance

    Args:
        distance: A float.

    Returns:
        An exponential score value for a given SMILES
    """
    return np.exp(-beta * distance**2)


class DistanceScorer(ScoringFunction):
    def __init__(self, beta: float = 0.00000001) -> None:
        """DistanceScorer is used to call a partial copy of distance_to_score function.

        Args:
            beta: A float value used for getting an exponential score value
        """
        self.partical_distance_score = partial(distance_to_score, beta=beta)

    def get_distance(self, smile_distance: float) -> float:
        """Generates a partial copy of distance_to_score function

        Args:
            smiles: SMILES.

        Returns:
            An exponential score value for a given SMILES
        """
        return self.partical_distance_score(smile_distance)


class TargetValueScorer(DistanceScorer):
    def __init__(self, target: float, scoring_function: Callable[[str], float]) -> None:
        """Scoring function which is used to generate a score based on a taget and a scoring function.

        Args:
            target: target score that will be used to get the distance to the score of the SMILES
            scoring_function: an instance of a scoring class
        """
        super().__init__()
        self.target = target
        self.scoring_function = scoring_function

    def score(self, smiles: str) -> float:
        """Generates a score for a given SMILES

        Args:
            smiles: SMILES.

        Returns:
            A score for the given SMILES
        """
        return self.get_distance(self.scoring_function(smiles) - self.target)

    def score_list(self, smiles_list: List[str]) -> List[float]:
        """Generates a list of scores for a given SMILES List

        Args:
            smiles_list: A List of SMILES.

        Returns:
            A List of scores
        """
        return [
            self.score(smiles)
            for smiles in smiles_list
            if Chem.MolFromSmiles(smiles) and smiles
        ]


class CombinedScorer:
    def __init__(
        self,
        scorer_list: List[Type[Any]],
        weights: Optional[List[float]] = None,
    ) -> None:
        """Scoring function which generates a combined score for a SMILES as per the given scoring functions.

        Args:
            scorer_list: A list of the scoring functions
            weights: A list of weights
        """
        self.scorer_list = scorer_list
        self.weights = self._normalize_weights(weights)

    def _normalize_weights(self, weights=None) -> List[float]:
        """It is used for normalizing weights.

        Args:
            weights: A list of weights.

        Returns:
            Sum of all the scores generated by the given scoring functions
        """
        weights = weights if weights else [1.0] * len(self.scorer_list)
        offsetted_weights = [weight + min(weights) for weight in weights]
        return [weight / float(sum(offsetted_weights)) for weight in offsetted_weights]

    def score(self, smiles: str):
        """Generates a score for a given SMILES

        Args:
            smiles: SMILES.

        Returns:
            Sum of all the scores generated by the given scoring functions
        """
        return sum(
            [
                scorer.score(smiles) * weight
                for scorer, weight in zip(self.scorer_list, self.weights)
            ]
        )

    def score_list(self, smiles_list: List[str]) -> List[float]:
        """Generates a list of scores for a given SMILES List

        Args:
            smiles_list: A List of SMILES.

        Returns:
            A List of scores
        """
        return [self.score(smiles) for smiles in smiles_list]


class RDKitDescriptorScorer(TargetValueScorer):
    def __init__(
        self,
        target: float,
        modifier: str = "gaussian_modifier",
        descriptor: str = "num_rotatable_bonds",
    ) -> None:
        """Scoring function wrapping RDKit descriptors.

        Args:
            target: target score that will be used to get the distance to the score of the SMILES
            modifier: score modifier
            descriptor:  molecular descriptors
        """
        self.target = target
        self.modifier = MODIFIERS[modifier](**MODIFIERS_PARAMETERS[modifier])
        self.descriptor = DESCRIPTOR[descriptor]
        super().__init__(target=target, scoring_function=self.score)

    def score(self, smiles: str) -> float:
        """Generates a score for a given SMILES

        Args:
            smiles: SMILES.

        Returns:
            A score for the given SMILES
        """
        scoring_function = RdkitScoringFunction(
            descriptor=self.descriptor,
            score_modifier=self.modifier,
        )
        return scoring_function.score_mol(Chem.MolFromSmiles(smiles))  # type: ignore


class TanimotoScorer(TargetValueScorer):
    def __init__(
        self,
        target: float,
        target_smile: str,
        fp_type: str = "ECFP4",
        modifier: str = "gaussian_modifier",
    ) -> None:
        """Scoring function that looks at the fingerprint similarity against a target molecule.

        Args:
            target: target score that will be used to get the distance to the score of the SMILES
            target_smile: target molecule to compare similarity
            fp_type: fingerprint type
            modifier: score modifier
        """
        self.target = target
        self.target_smile = target_smile
        self.fp_type = fp_type
        self.modifier = MODIFIERS[modifier](**MODIFIERS_PARAMETERS[modifier])
        super().__init__(target=target, scoring_function=self.score)

    def score(self, smiles: str) -> float:
        """Generates a score for a given SMILES

        Args:
            smiles: SMILES.

        Returns:
            A score for the given SMILES
        """
        scoring_function = TanimotoScoringFunction(
            self.target_smile,
            fp_type=self.fp_type,
            score_modifier=self.modifier,
        )
        return scoring_function.score_mol(Chem.MolFromSmiles(smiles))  # type: ignore


class IsomerScorer(TargetValueScorer):
    def __init__(self, target: float, target_smile: str) -> None:
        """Scoring function for closeness to a molecular formula.

        Args:
            target: target score that will be used to get the distance to the score of the SMILES
            target_smile: targeted SMILES to compare closeness with
        """
        self.target = target
        self.target_smile = target_smile
        super().__init__(target=target, scoring_function=self.score)

    def score(self, smiles: str) -> float:
        """Generates a score for a given SMILES

        Args:
            smiles: SMILES.

        Returns:
            A score for the given SMILES
        """
        scoring_function = IsomerScoringFunction(self.target_smile)
        return scoring_function.raw_score(smiles)


class SMARTSScorer(TargetValueScorer):
    def __init__(self, target: float, target_smile: str, inverse: bool = True) -> None:
        """Scoring function that looks at the fingerprint similarity against a target molecule.

        Args:
            target: target score that will be used to get the distance to the score of the SMILES
            target_smile: The SMARTS string to match
            inverse: If True then SMARTS is desired else it is not desired in the molecules
        """
        self.target = target
        self.target_smile = target_smile
        self.inverse = inverse
        super().__init__(target=target, scoring_function=self.score)

    def score(self, smiles: str) -> float:
        """Generates a score for a given SMILES

        Args:
            smiles: SMILES.

        Returns:
            A score for the given SMILES
        """
        scoring_function = SMARTSScoringFunction(self.target_smile, self.inverse)
        return scoring_function.score_mol(Chem.MolFromSmiles(smiles))  # type: ignore


class QEDScorer(TargetValueScorer):
    def __init__(self, target: float) -> None:
        """Scoring function that calculates the weighted sum of ADS mapped properties using QED module of rdkit

        Args:
            target: target score that will be used to get the distance to the score of the SMILES
        """
        self.target = target
        super().__init__(target=target, scoring_function=self.score)

    def score(self, smiles: str) -> float:
        """Generates a score for a given SMILES

        Args:
            smiles: SMILES.

        Returns:
            A score for the given SMILES
        """
        return Chem.QED.qed(Chem.MolFromSmiles(smiles))



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""S3 storage tests.

These assume a bucket `gt4sd-ci-tests` and an artifact `a_folder/containing/a_file.txt`.
"""

import logging
import os
import shutil
import tempfile

import pytest

from gt4sd.s3 import GT4SDS3Client, sync_folder_with_s3
from gt4sd.tests.utils import GT4SDTestSettings

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

test_settings = GT4SDTestSettings.get_instance()


@pytest.fixture
def client() -> GT4SDS3Client:
    return GT4SDS3Client(
        host=test_settings.gt4sd_s3_host,
        access_key=test_settings.gt4sd_s3_access_key,
        secret_key=test_settings.gt4sd_s3_secret_key,
        secure=test_settings.gt4sd_s3_secure,
    )


def test_bucket_listing(client: GT4SDS3Client):
    bucket_names = client.list_bucket_names()
    assert isinstance(bucket_names, list)
    assert "gt4sd-ci-tests" in bucket_names


def test_object_listing(client: GT4SDS3Client):
    object_names = client.list_object_names(bucket="gt4sd-ci-tests", prefix="a_folder")
    assert isinstance(object_names, list)
    for object_name in object_names:
        assert object_name.endswith(".txt")


def test_directory_listing(client: GT4SDS3Client):
    object_names = client.list_directories(bucket="gt4sd-ci-tests", prefix="a_folder")
    assert isinstance(object_names, set)
    for object_name in object_names:
        assert object_name == "containing"


def test_sync_folder(client: GT4SDS3Client):
    directory = tempfile.mkdtemp()
    try:
        client.sync_folder("gt4sd-ci-tests", directory)
        filepaths = set(
            [filepath.replace(directory, "") for filepath in os.listdir(directory)]
        )
        objectpaths = set(
            map(
                lambda path: path.split("/")[0],
                client.list_object_names(bucket="gt4sd-ci-tests"),
            )
        )
        assert filepaths == objectpaths
    finally:
        logger.info(f"cleaning up test folder {directory}")
        shutil.rmtree(directory)


def test_sync_folder_with_s3(client):
    directory = tempfile.mkdtemp()
    try:
        sync_folder_with_s3(
            host=test_settings.gt4sd_s3_host,
            access_key=test_settings.gt4sd_s3_access_key,
            secret_key=test_settings.gt4sd_s3_secret_key,
            bucket="gt4sd-ci-tests",
            folder_path=directory,
            secure=test_settings.gt4sd_s3_secure,
        )
        filepaths = set(
            [filepath.replace(directory, "") for filepath in os.listdir(directory)]
        )
        objectpaths = set(
            map(
                lambda path: path.split("/")[0],
                client.list_object_names(bucket="gt4sd-ci-tests"),
            )
        )
        assert filepaths == objectpaths
    finally:
        logger.info(f"cleaning up test folder {directory}")
        shutil.rmtree(directory)



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""GFlowNet training utilities."""

import logging
from dataclasses import dataclass, field
from typing import Any, Dict, Optional, Tuple, Union

import sentencepiece as _sentencepiece
import torch as _torch
import tensorflow as _tensorflow
from gt4sd_trainer.hf_pl.pytorch_lightning_trainer import (
    PytorchLightningTrainingArguments,
    PyTorchLightningTrainingPipeline,
)
from pytorch_lightning import LightningDataModule, LightningModule, Trainer
from pytorch_lightning.loggers import TensorBoardLogger

from ....frameworks.gflownet.dataloader.data_module import (
    GFlowNetDataModule,
    GFlowNetTask,
)
from ....frameworks.gflownet.dataloader.dataset import GFlowNetDataset
from ....frameworks.gflownet.envs.graph_building_env import (
    GraphBuildingEnv,
    GraphBuildingEnvContext,
)
from ....frameworks.gflownet.loss import ALGORITHM_FACTORY
from ....frameworks.gflownet.ml.models import MODEL_FACTORY
from ....frameworks.gflownet.ml.module import GFlowNetModule
from ...core import TrainingPipelineArguments

# imports that have to be loaded before lightning to avoid segfaults
_sentencepiece
_tensorflow
_torch

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class GFlowNetTrainingPipeline(PyTorchLightningTrainingPipeline):
    """gflownet training pipelines."""

    def train(  # type: ignore
        self,
        pl_trainer_args: Dict[str, Any],
        model_args: Dict[str, Union[float, str, int]],
        dataset_args: Dict[str, Union[float, str, int]],
        dataset: GFlowNetDataset,
        environment: GraphBuildingEnv,
        context: GraphBuildingEnvContext,
        task: GFlowNetTask,
    ) -> None:
        """Generic training function for PyTorch Lightning-based training.

        Args:
            pl_trainer_args: pytorch lightning trainer arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.
            dataset: dataset to be used for training.
            environment: environment to be used for training.
            context: context to be used for training.
            task: task to be used for training.
        """

        logger.info(f"Trainer arguments: {pl_trainer_args}")

        if pl_trainer_args[
            "resume_from_checkpoint"
        ] is not None and not pl_trainer_args["resume_from_checkpoint"].endswith(
            ".ckpt"
        ):
            pl_trainer_args["resume_from_checkpoint"] = None

        pl_trainer_args["callbacks"] = {
            "model_checkpoint_callback": {"save_top_k": pl_trainer_args["save_top_k"]}
        }

        pl_trainer_args["callbacks"] = self.add_callbacks(pl_trainer_args["callbacks"])

        pl_trainer_args["logger"] = TensorBoardLogger(
            pl_trainer_args["save_dir"], name=pl_trainer_args["basename"]
        )

        trainer = Trainer(
            profiler=pl_trainer_args["profiler"],
            logger=pl_trainer_args["logger"],
            log_every_n_steps=pl_trainer_args["trainer_log_every_n_steps"],
            callbacks=pl_trainer_args["callbacks"],
            max_epochs=pl_trainer_args["epochs"],
            strategy=pl_trainer_args["strategy"],
            fast_dev_run=pl_trainer_args["development_mode"],
        )

        data_module, model_module = self.get_data_and_model_modules(
            model_args,
            dataset_args,
            pl_trainer_args,
            dataset,
            environment,
            context,
            task,
        )
        trainer.fit(model_module, data_module)

    # TODO: compatible signature with Pytorch Lightning training pipelines
    def get_data_and_model_modules(  # type: ignore
        self,
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
        pl_trainer_args: Dict[str, Any],
        dataset: GFlowNetDataset,
        environment: GraphBuildingEnv,
        context: GraphBuildingEnvContext,
        task: GFlowNetTask,
    ) -> Tuple[LightningDataModule, LightningModule]:
        """Get data and model modules for training.

        Args:
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Returns:
            the data and model modules.
        """

        configuration = {**model_args, **dataset_args, **pl_trainer_args}
        if configuration["algorithm"] in ALGORITHM_FACTORY:
            algorithm = ALGORITHM_FACTORY[configuration["algorithm"]](
                configuration,
                environment,
                context,
            )
        else:
            raise ValueError(
                "Algorithm configuration is not given in the specified config file."
            )

        if configuration["model"] in MODEL_FACTORY:
            model = MODEL_FACTORY[configuration["model"]](
                configuration,
                context,
            )
        else:
            raise ValueError(
                "Models configuration is not given in the specified config file."
            )

        dm = GFlowNetDataModule(
            configuration=configuration,
            dataset=dataset,
            environment=environment,
            context=context,
            task=task,
            algorithm=algorithm,
            model=model,
        )
        dm.prepare_data()

        module = GFlowNetModule(
            configuration=configuration,
            dataset=dataset,
            environment=environment,
            context=context,
            task=task,
            algorithm=algorithm,
            model=model,
        )

        return dm, module


@dataclass
class GFlowNetPytorchLightningTrainingArguments(PytorchLightningTrainingArguments):
    """
    Arguments related to pytorch lightning trainer.
    """

    __name__ = "pl_trainer_args"

    strategy: Optional[str] = field(
        default="ddp", metadata={"help": "Training strategy."}
    )
    accumulate_grad_batches: int = field(
        default=1,
        metadata={
            "help": "Accumulates grads every k batches or as set up in the dict."
        },
    )

    trainer_log_every_n_steps: int = field(
        default=50,
        metadata={"help": "log every k steps."},
    )

    val_check_interval: int = field(
        default=5000, metadata={"help": " How often to check the validation set."}
    )
    save_dir: Optional[str] = field(
        default="logs", metadata={"help": "Save directory for logs and output."}
    )
    basename: Optional[str] = field(
        default="lightning_logs", metadata={"help": "Experiment name."}
    )
    gradient_clip_val: float = field(
        default=0.0, metadata={"help": "Gradient clipping value."}
    )
    limit_val_batches: int = field(
        default=500, metadata={"help": "How much of validation dataset to check."}
    )
    log_every_n_steps: int = field(
        default=500, metadata={"help": "How often to log within steps."}
    )
    max_epochs: int = field(
        default=3,
        metadata={"help": "Stop training once this number of epochs is reached."},
    )
    epochs: int = field(
        default=3,
        metadata={"help": "Stop training once this number of epochs is reached."},
    )
    resume_from_checkpoint: Optional[str] = field(
        default=None,
        metadata={"help": "Path/URL of the checkpoint from which training is resumed."},
    )
    gpus: Optional[int] = field(
        default=-1,
        metadata={"help": "Number of gpus to train on."},
    )

    monitor: Optional[str] = field(
        default=None,
        metadata={"help": "Quantity to monitor in order to store a checkpoint."},
    )
    save_last: Optional[bool] = field(
        default=None,
        metadata={
            "help": "When True, always saves the model at the end of the epoch to a file last.ckpt"
        },
    )
    save_top_k: int = field(
        default=1,
        metadata={
            "help": "The best k models according to the quantity monitored will be saved."
        },
    )
    mode: str = field(
        default="min",
        metadata={"help": "Quantity to monitor in order to store a checkpoint."},
    )
    every_n_train_steps: Optional[int] = field(
        default=None,
        metadata={"help": "Number of training steps between checkpoints."},
    )
    check_val_every_n_epoch: Optional[int] = field(
        default=5,
        metadata={"help": "Number of validation epochs between checkpoints."},
    )
    auto_lr_find: bool = field(
        default=True,
        metadata={
            "help": "Select whether to run a learning rate finder to try to optimize initial learning for faster convergence."
        },
    )
    profiler: Optional[str] = field(
        default="simple",
        metadata={
            "help": "To profile individual steps during training and assist in identifying bottlenecks."
        },
    )
    learning_rate: float = field(
        default=0.0001,
        metadata={"help": "The learning rate."},
    )

    test_output_path: Optional[str] = field(
        default="./test",
        metadata={
            "help": "Path where to save latent encodings and predictions for the test set when an epoch ends."
        },
    )
    num_workers: int = field(
        default=0,
        metadata={"help": "number of workers. Defaults to 1."},
    )

    log_dir: str = field(
        default="./log/",
        metadata={"help": "The directory to save logs."},
    )

    num_training_steps: int = field(
        default=1000,
        metadata={"help": "The number of training steps."},
    )

    validate_every: int = field(
        default=1000,
        metadata={"help": "The number of training steps between validation."},
    )

    seed: int = field(
        default=142857,
        metadata={"help": "The random seed."},
    )

    device: str = field(
        default="cpu",
        metadata={"help": "The device to use."},
    )
    development_mode: bool = field(
        default=False,
        metadata={"help": "Whether to run in development mode. "},
    )


@dataclass
class GFlowNetModelArguments(TrainingPipelineArguments):
    """
    Arguments related to model.
    """

    __name__ = "model_args"

    algorithm: str = field(
        default="trajectory_balance",
        metadata={"help": "The algorithm to use for training the model. "},
    )
    context: str = field(  # type: ignore
        default=None,
        metadata={"help": "The environment context to use for training the model. "},
    )
    environment: str = field(  # type: ignore
        default=None,
        metadata={"help": "The environment to use for training the model. "},
    )
    model: str = field(
        default="graph_transformer_gfn",
        metadata={"help": "The model to use for training the model. "},
    )
    sampling_model: str = field(
        default="graph_transformer_gfn",
        metadata={"help": "The model used to generate samples. "},
    )
    task: str = field(
        default="qm9",
        metadata={"help": "The task to use for training the model. "},
    )

    bootstrap_own_reward: bool = field(
        default=False,
        metadata={"help": "Whether to bootstrap the own reward. "},
    )

    num_emb: int = field(
        default=128,
        metadata={"help": "The number of embeddings. "},
    )

    num_layers: int = field(
        default=4,
        metadata={"help": "The number of layers. "},
    )

    tb_epsilon: float = field(
        default=1e-10,
        metadata={"help": "The epsilon. "},
    )

    illegal_action_logreward: float = field(
        default=-50.0,
        metadata={"help": "The illegal action log reward. "},
    )

    reward_loss_multiplier: float = field(
        default=1.0,
        metadata={"help": "The reward loss multiplier. "},
    )

    temperature_sample_dist: str = field(
        default="uniform",
        metadata={"help": "The temperature sample distribution. "},
    )

    temperature_dist_params: str = field(
        default="(.5, 32)",
        metadata={"help": "The temperature distribution parameters. "},
    )

    weight_decay: float = field(
        default=1e-8,
        metadata={"help": "The weight decay. "},
    )
    momentum: float = field(
        default=0.9,
        metadata={"help": "The momentum. "},
    )

    adam_eps: float = field(
        default=1e-8,
        metadata={"help": "The adam epsilon. "},
    )

    lr_decay: float = field(
        default=20000,
        metadata={"help": "The learning rate decay steps. "},
    )

    z_lr_decay: float = field(
        default=20000,
        metadata={"help": "The learning rate decay steps for z."},
    )

    clip_grad_type: str = field(
        default="norm",
        metadata={"help": "The clip grad type. "},
    )

    clip_grad_param: float = field(
        default=10.0,
        metadata={"help": "The clip grad param. "},
    )

    random_action_prob: float = field(
        default=0.001,
        metadata={"help": "The random action probability. "},
    )

    sampling_tau: float = field(
        default=0.0,
        metadata={"help": "The sampling temperature. "},
    )

    max_nodes: int = field(
        default=9,
        metadata={"help": "The maximum number of nodes. "},
    )

    num_offline: int = field(
        default=10,
        metadata={"help": "The number of offline samples. "},
    )


@dataclass
class GFlowNetDataArguments(TrainingPipelineArguments):
    """
    Arguments related to data.
    """

    __name__ = "dataset_args"

    dataset: str = field(
        default="qm9",
        metadata={"help": "The dataset to use for training the model. "},
    )
    dataset_path: str = field(
        default="./data/qm9",
        metadata={"help": "The path to the dataset to use for training the model. "},
    )

    batch_size: int = field(
        default=64,
        metadata={"help": "Batch size of the training. Defaults to 64."},
    )
    global_batch_size: int = field(
        default=16,
        metadata={"help": "Global batch size of the training. Defaults to 16."},
    )

    validation_split: Optional[float] = field(
        default=None,
        metadata={
            "help": "Proportion used for validation. Defaults to None, a.k.a., use indices file if provided otherwise uses half of the data for validation."
        },
    )
    validation_indices_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "Indices to use for validation. Defaults to None, a.k.a., use validation split proportion, if not provided uses half of the data for validation."
        },
    )
    stratified_batch_file: Optional[str] = field(
        default=None,
        metadata={
            "help": "Stratified batch file for sampling. Defaults to None, a.k.a., no stratified sampling."
        },
    )
    stratified_value_name: Optional[str] = field(
        default=None,
        metadata={
            "help": "Stratified value name. Defaults to None, a.k.a., no stratified sampling. Needed in case a stratified batch file is provided."
        },
    )
    num_data_loader_workers: int = field(
        default=8,
        metadata={"help": "The number of data loader workers. "},
    )
    sampling_iterator: bool = field(
        default=True,
        metadata={"help": "Whether to use a sampling iterator. "},
    )

    ratio: float = field(
        default=0.9,
        metadata={"help": "The ratio. "},
    )


@dataclass
class GFlowNetSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to Granular trainer."""

    __name__ = "saving_args"

    model_path: str = field(
        metadata={"help": "Path to the checkpoint file to be used."}
    )



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Controlled Latent attribute Space Sampling initialization."""
import logging

from ....extras import EXTRAS_ENABLED

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

if EXTRAS_ENABLED:
    from .core import PAG, CLaSS, CogMol

    __all__ = ["CLaSS", "CogMol", "PAG"]
else:
    logger.warning("install cogmol-inference extras to use CLaSS")



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""MoLeR tests."""

from typing import ClassVar, Type

import pytest

from gt4sd.algorithms.core import AlgorithmConfiguration
from gt4sd.algorithms.generation.moler import MoLeR, MoLeRDefaultGenerator
from gt4sd.algorithms.registry import ApplicationsRegistry


def get_classvar_type(class_var):
    """Extract type from ClassVar type annotation: `ClassVar[T]] -> T`."""
    return class_var.__args__[0]


@pytest.mark.parametrize(
    "config_class, algorithm_type, domain, algorithm_name",
    [
        (
            MoLeRDefaultGenerator,
            "generation",
            "materials",
            MoLeR.__name__,
        )
    ],
)
def test_config_class(
    config_class: Type[AlgorithmConfiguration],
    algorithm_type: str,
    domain: str,
    algorithm_name: str,
):
    assert config_class.algorithm_type == algorithm_type
    assert config_class.domain == domain
    assert config_class.algorithm_name == algorithm_name

    for keyword, type_annotation in config_class.__annotations__.items():
        if keyword in ("algorithm_type", "domain", "algorithm_name"):
            assert type_annotation.__origin__ is ClassVar  # type: ignore
            assert str == get_classvar_type(type_annotation)


@pytest.mark.parametrize(
    "config_class",
    [(MoLeRDefaultGenerator)],
)
def test_config_instance(config_class: Type[AlgorithmConfiguration]):
    config = config_class()  # type:ignore
    assert config.algorithm_application == config_class.__name__


@pytest.mark.parametrize(
    "config_class",
    [(MoLeRDefaultGenerator)],
)
def test_available_versions(config_class: Type[AlgorithmConfiguration]):
    versions = config_class.list_versions()
    assert "v0" in versions


@pytest.mark.parametrize(
    "config, algorithm, params",
    [
        # Test unconditional generation
        (MoLeRDefaultGenerator, MoLeR, {"seed": 42, "num_workers": 1}),
        # Test soft-conditioning on latent codes
        (
            MoLeRDefaultGenerator,
            MoLeR,
            {
                "seed": 0,
                "seed_smiles": "c1ccccc1.CNC=O.CN1C=NC2=C1C(=O)N(C(=O)N2C)C.[O-]C(=O)[O-].CCOCN",
                "num_workers": 1,
            },
        ),
        # Test hard-conditioning on scaffolds
        (
            MoLeRDefaultGenerator,
            MoLeR,
            {
                "seed": 0,
                "scaffolds": "CN.O=C1C2=CC=C(C3=CC=CC=C3)C=C=C2OC2=CC=CC=C12",
                "num_workers": 1,
            },
        ),
        # Test both together
        (
            MoLeRDefaultGenerator,
            MoLeR,
            {
                "seed": 0,
                "seed_smiles": "c1ccccc1.CNC=O.CN1C=NC2=C1C(=O)N(C(=O)N2C)C.[O-]C(=O)[O-].CCOCN",
                "scaffolds": "CN.CCC",
                "num_workers": 1,
            },
        ),
        # Test both together with unequal sizes
        (
            MoLeRDefaultGenerator,
            MoLeR,
            {
                "seed": 0,
                "seed_smiles": "c1ccccc1.CNC=O.CN1C=NC2=C1C(=O)N(C(=O)N2C)C.[O-]C(=O)[O-].CCOCN",
                "scaffolds": "CN.CCC.CNO",
                "num_workers": 1,
            },
        ),
    ],
)
def test_generation_via_import(config, algorithm, params):
    algorithm = algorithm(configuration=config(**params))
    items = list(algorithm.sample(5))
    assert len(items) == 5


@pytest.mark.parametrize(
    "algorithm_application, algorithm_type, domain, algorithm_name",
    [
        (
            MoLeRDefaultGenerator.__name__,
            "generation",
            "materials",
            MoLeR.__name__,
        ),
    ],
)
def test_generation_via_registry(
    algorithm_type, domain, algorithm_name, algorithm_application
):
    algorithm = ApplicationsRegistry.get_application_instance(
        target=None,
        algorithm_type=algorithm_type,
        domain=domain,
        algorithm_name=algorithm_name,
        algorithm_application=algorithm_application,
    )
    items = list(algorithm.sample(5))
    assert len(items) == 5



#
# MIT License
#
# Copyright (c) 2022 GT4SD team
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""Diffusion training utilities. Code adapted from: https://github.com/huggingface/diffusers/blob/main/examples/unconditional_image_generation/train_unconditional.py"""

import logging
import os
from dataclasses import dataclass, field
from typing import Any, Dict

import torch
import torch.nn.functional as F
from accelerate import Accelerator
from datasets import load_dataset
from diffusers import DDPMPipeline, DDPMScheduler, UNet2DModel
from diffusers.optimization import get_scheduler
from diffusers.training_utils import EMAModel
from rdkit.Chem import Descriptors
from torchvision.transforms import (
    CenterCrop,
    Compose,
    InterpolationMode,
    Normalize,
    RandomHorizontalFlip,
    Resize,
    ToTensor,
)
from tqdm.auto import tqdm

from ..core import TrainingPipeline, TrainingPipelineArguments

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# NOTE: This import is needed because importing torchvision before rdkit Descriptors
# can cause segmentation faults.
Descriptors


class DiffusionForVisionTrainingPipeline(TrainingPipeline):
    """Diffusion training pipelines for image generation."""

    def train(  # type: ignore
        self,
        training_args: Dict[str, Any],
        model_args: Dict[str, Any],
        dataset_args: Dict[str, Any],
    ) -> None:
        """Generic training function for Diffusion models.

        Args:
            training_args: training arguments passed to the configuration.
            model_args: model arguments passed to the configuration.
            dataset_args: dataset arguments passed to the configuration.

        Raises:
            NotImplementedError: the generic trainer does not implement the pipeline.
        """
        params = {**training_args, **dataset_args, **model_args}

        env_local_rank = int(os.environ.get("LOCAL_RANK", -1))
        if env_local_rank != -1 and env_local_rank != params["local_rank"]:
            params["local_rank"] = env_local_rank

        logging_dir = os.path.join(params["output_dir"], params["logging_dir"])
        model_path = params["model_path"]
        training_name = params["training_name"]

        accelerator = Accelerator(
            mixed_precision=params["mixed_precision"],
            log_with="tensorboard",
            logging_dir=logging_dir,
        )
        logger.info(f"Model with name {training_name} starts.")

        model_dir = os.path.join(model_path, training_name)
        log_path = os.path.join(model_dir, "logs")
        val_dir = os.path.join(log_path, "val_logs")

        os.makedirs(os.path.join(model_dir, "weights"), exist_ok=True)
        os.makedirs(os.path.join(model_dir, "results"), exist_ok=True)
        os.makedirs(log_path, exist_ok=True)
        os.makedirs(val_dir, exist_ok=True)

        # unet decoder
        model = UNet2DModel(
            sample_size=params["resolution"],
            in_channels=params["in_channels"],
            out_channels=params["out_channels"],
            layers_per_block=params["layers_per_block"],
            block_out_channels=(128, 128, 256, 256, 512, 512),
            down_block_types=(
                "DownBlock2D",
                "DownBlock2D",
                "DownBlock2D",
                "DownBlock2D",
                "AttnDownBlock2D",
                "DownBlock2D",
            ),
            up_block_types=(
                "UpBlock2D",
                "AttnUpBlock2D",
                "UpBlock2D",
                "UpBlock2D",
                "UpBlock2D",
                "UpBlock2D",
            ),
        )
        # ddpm noise schedule
        noise_scheduler = DDPMScheduler(
            num_train_timesteps=params["num_train_timesteps"]
        )
        optimizer = torch.optim.AdamW(
            model.parameters(),
            lr=params["learning_rate"],
            betas=(params["adam_beta1"], params["adam_beta2"]),
            weight_decay=params["adam_weight_decay"],
            eps=params["adam_epsilon"],
        )

        augmentations = Compose(
            [
                Resize(params["resolution"], interpolation=InterpolationMode.BILINEAR),
                CenterCrop(params["resolution"]),
                RandomHorizontalFlip(),
                ToTensor(),
                Normalize([0.5], [0.5]),
            ]
        )
        # specify dataset by name or path
        if params["dataset_name"] is not None:
            dataset = load_dataset(
                params["dataset_name"],
                params["dataset_config_name"],
                cache_dir=params["cache_dir"],
                use_auth_token=True if params["use_auth_token"] else None,
                split="train",
            )
            logger.info("dataset name: " + params["dataset_name"])
        else:
            if params["train_data_dir"] is None:
                raise ValueError(
                    "You must specify either a dataset name from the hub or a train data directory."
                )
            dataset = load_dataset(
                "imagefolder",
                data_dir=params["train_data_dir"],
                cache_dir=params["cache_dir"],
                split="train",
            )
            logger.info("dataset path: " + params["train_data_dir"])

        def transforms(examples):
            try:
                images = [
                    augmentations(image.convert("RGB")) for image in examples["img"]
                ]
            except KeyError:
                images = [
                    augmentations(image.convert("RGB")) for image in examples["image"]
                ]
            return {"input": images}

        dataset.set_transform(transforms)  # type: ignore
        train_dataloader = torch.utils.data.DataLoader(  # type: ignore
            dataset, batch_size=params["train_batch_size"], shuffle=True  # type: ignore
        )

        # specify learning rate scheduler
        lr_scheduler = get_scheduler(
            params["lr_scheduler"],
            optimizer=optimizer,
            num_warmup_steps=params["lr_warmup_steps"],
            num_training_steps=(len(train_dataloader) * params["num_epochs"])
            // params["gradient_accumulation_steps"],
        )

        # preparare for distributed training if neeeded
        model, optimizer, train_dataloader, lr_scheduler = accelerator.prepare(
            model, optimizer, train_dataloader, lr_scheduler
        )

        # initialize the ema model
        ema_model = EMAModel(
            model,
            inv_gamma=params["ema_inv_gamma"],
            power=params["ema_power"],
            max_value=params["ema_max_decay"],
        )

        if accelerator.is_main_process:
            run = os.path.split(__file__)[-1].split(".")[0]
            accelerator.init_trackers(run)

        global_step = 0
        # start training
        for epoch in range(params["num_epochs"]):

            model.train()
            # progress bar visualization
            progress_bar = tqdm(
                total=len(train_dataloader),
                disable=not accelerator.is_local_main_process,
            )
            progress_bar.set_description(f"Epoch {epoch}")

            for _, batch in enumerate(train_dataloader):

                clean_images = batch["input"]
                # Sample noise that we'll add to the images
                noise = torch.randn(clean_images.shape).to(clean_images.device)
                bsz = clean_images.shape[0]
                # Sample a random timestep for each image
                timesteps = torch.randint(
                    0,
                    noise_scheduler.num_train_timesteps,
                    (bsz,),
                    device=clean_images.device,
                ).long()

                # Add noise to the clean images according to the noise magnitude at each timestep
                # (this is the forward diffusion process)
                noisy_images = noise_scheduler.add_noise(clean_images, noise, timesteps)

                with accelerator.accumulate(model):
                    # Predict the noise residual
                    noise_pred = model(noisy_images, timesteps)["sample"]
                    loss = F.mse_loss(noise_pred, noise)
                    accelerator.backward(loss)

                    accelerator.clip_grad_norm_(model.parameters(), 1.0)
                    optimizer.step()
                    lr_scheduler.step()
                    if params["use_ema"]:
                        ema_model.step(model)
                    optimizer.zero_grad()

                progress_bar.update(1)
                logs = {
                    "loss": loss.detach().item(),
                    "lr": lr_scheduler.get_last_lr()[0],
                    "step": global_step,
                }
                if params["use_ema"]:
                    logs["ema_decay"] = ema_model.decay
                progress_bar.set_postfix(**logs)

                accelerator.log(logs, step=global_step)
                global_step += 1

                if params["dummy_training"]:
                    break

            progress_bar.close()
            # wait for all the processes to finish
            accelerator.wait_for_everyone()

            # Generate sample images for visual inspection
            if accelerator.is_main_process and params["is_sampling"]:
                if (
                    epoch % params["save_images_epochs"] == 0
                    or epoch == params["num_epochs"] - 1
                ):
                    # inference/sampling
                    pipeline = DDPMPipeline(
                        unet=accelerator.unwrap_model(
                            ema_model.averaged_model if params["use_ema"] else model
                        ),
                        scheduler=noise_scheduler,
                    )

                    generator = torch.manual_seed(0)
                    # run pipeline in inference (sample random noise and denoise)
                    images = pipeline(
                        generator=generator,
                        batch_size=params["eval_batch_size"],
                        output_type="numpy",
                    )["sample"]

                    # denormalize the images and save to tensorboard
                    images_processed = (images * 255).round().astype("uint8")
                    accelerator.trackers[0].writer.add_images(
                        "test_samples",
                        images_processed.transpose(0, 3, 1, 2),
                        epoch,
                    )

                if (
                    epoch % params["save_model_epochs"] == 0
                    or epoch == params["num_epochs"] - 1
                ):
                    pipeline.save_pretrained(params["output_dir"])
            accelerator.wait_for_everyone()

            if params["dummy_training"]:
                break

        accelerator.end_training()
        logger.info("Training done, shutting down.")


@dataclass
class DiffusionDataArguments(TrainingPipelineArguments):
    """Data arguments related to diffusion trainer."""

    __name__ = "dataset_args"

    dataset_name: str = field(default="", metadata={"help": "Dataset name."})
    dataset_config_name: str = field(
        default="", metadata={"help": "Dataset config name."}
    )
    train_data_dir: str = field(default="", metadata={"help": "Train data directory."})
    resolution: int = field(default=64, metadata={"help": "Resolution."})
    train_batch_size: int = field(default=16, metadata={"help": "Train batch size."})
    eval_batch_size: int = field(default=16, metadata={"help": "Eval batch size."})
    num_epochs: int = field(default=100, metadata={"help": "Number of epochs."})


@dataclass
class DiffusionModelArguments(TrainingPipelineArguments):
    """Model arguments related to Diffusion trainer."""

    __name__ = "model_args"

    model_path: str = field(metadata={"help": "Path to the model file."})
    training_name: str = field(metadata={"help": "Name of the training run."})

    num_train_timesteps: int = field(
        default=1000, metadata={"help": "Number of noise steps."}
    )
    learning_rate: float = field(default=1e-4, metadata={"help": "Learning rate."})
    lr_scheduler: str = field(
        default="cosine", metadata={"help": "Learning rate scheduler."}
    )
    lr_warm_up_steps: int = field(
        default=500, metadata={"help": "Learning rate warm up steps."}
    )
    adam_beta1: float = field(default=0.95, metadata={"help": "Adam beta1."})
    adam_beta2: float = field(default=0.999, metadata={"help": "Adam beta2."})
    adam_weight_decay: float = field(
        default=1e-6, metadata={"help": "Adam weights decay."}
    )
    adam_epsilon: float = field(default=1e-8, metadata={"help": "Adam eps."})
    gradient_accumulation_steps: int = field(
        default=1, metadata={"help": "Gradient accumulation steps."}
    )
    in_channels: int = field(default=3, metadata={"help": "Input channels."})
    out_channels: int = field(default=3, metadata={"help": "Output channels."})
    layers_per_block: int = field(default=2, metadata={"help": "Layers per block."})


@dataclass
class DiffusionTrainingArguments(TrainingPipelineArguments):
    """Training arguments related to Diffusion trainer."""

    __name__ = "training_args"

    local_rank: int = field(default=-1, metadata={"help": "Local rank of the process."})
    output_dir: str = field(
        default="ddpm-cifar10-32", metadata={"help": "Output directory."}
    )
    logging_dir: str = field(default="logs/", metadata={"help": "Logging directory."})
    overwrite_output_dir: bool = field(
        default=False, metadata={"help": "Overwrite output directory."}
    )
    cache_dir: str = field(default=".cache/", metadata={"help": "Cache directory."})
    save_images_epochs: int = field(
        default=10, metadata={"help": "Save images every n epochs."}
    )
    save_model_epochs: int = field(
        default=10, metadata={"help": "Save model every n epochs."}
    )
    use_ema: bool = field(default=True, metadata={"help": "Use ema."})
    ema_inv_gamma: float = field(default=1.0, metadata={"help": "Ema inverse gamma."})
    ema_power: float = field(default=0.75, metadata={"help": "Ema power."})
    ema_max_decay: float = field(default=0.9999, metadata={"help": "Ema max delay."})
    mixed_precision: str = field(
        default="no",
        metadata={"help": "Mixed precision. Choose from 'no', 'fp16', 'bf16'."},
    )
    use_auth_token: bool = field(
        default=False,
        metadata={
            "help": "Use the token generated when using huggingface-hub (necessary to use this script with private models)."
        },
    )
    dummy_training: bool = field(
        default=False,
        metadata={"help": "Run dummy training to test the pipeline."},
    )
    is_sampling: bool = field(
        default=True,
        metadata={"help": "Run sampling."},
    )


@dataclass
class DiffusionSavingArguments(TrainingPipelineArguments):
    """Saving arguments related to Diffusion trainer."""

    __name__ = "saving_args"
